from __future__ import division
import numpy as np
import scipy.sparse as sp
from scipy.constants import epsilon_0

import SimPEG
from SimPEG import Utils
from SimPEG.EM.Utils import omega
from SimPEG.Utils import Zero, sdiag


class FieldsTDEM(SimPEG.Problem.TimeFields):
    """

    Fancy Field Storage for a TDEM survey. Only one field type is stored for
    each problem, the rest are computed. The fields obejct acts like an array
    and is indexed by

    .. code-block:: python

        f = problem.fields(m)
        e = f[srcList,'e']
        b = f[srcList,'b']

    If accessing all sources for a given field, use the :code:`:`

    .. code-block:: python

        f = problem.fields(m)
        e = f[:,'e']
        b = f[:,'b']

    The array returned will be size (nE or nF, nSrcs :math:`\\times`
    nFrequencies)
    """

    knownFields = {}
    dtype = float

    def _GLoc(self, fieldType):
        """Grid location of the fieldType"""
        return self.aliasFields[fieldType][1]

    def _eDeriv(self, tInd, src, dun_dm_v, v, adjoint=False):
        if adjoint is True:
            return (
                self._eDeriv_u(tInd, src, v, adjoint),
                self._eDeriv_m(tInd, src, v, adjoint)
            )
        return (
            self._eDeriv_u(tInd, src, dun_dm_v) +
            self._eDeriv_m(tInd, src, v)
        )

    def _bDeriv(self, tInd, src, dun_dm_v, v, adjoint=False):
        if adjoint is True:
            return (
                self._bDeriv_u(tInd, src, v, adjoint),
                self._bDeriv_m(tInd, src, v, adjoint)
            )
        return (
            self._bDeriv_u(tInd, src, dun_dm_v) +
            self._bDeriv_m(tInd, src, v)
        )

    def _dbdtDeriv(self, tInd, src, dun_dm_v, v, adjoint=False):
        if adjoint is True:
            return (
                self._dbdtDeriv_u(tInd, src, v, adjoint),
                self._dbdtDeriv_m(tInd, src, v, adjoint)
            )
        return (
            self._dbdtDeriv_u(tInd, src, dun_dm_v) +
            self._dbdtDeriv_m(tInd, src, v)
        )

    def _hDeriv(self, tInd, src, dun_dm_v, v, adjoint=False):
        if adjoint is True:
            return (
                self._hDeriv_u(tInd, src, v, adjoint),
                self._hDeriv_m(tInd, src, v, adjoint)
            )
        return (
            self._hDeriv_u(tInd, src, dun_dm_v) +
            self._hDeriv_m(tInd, src, v)
        )

    def _dhdtDeriv(self, tInd, src, dun_dm_v, v, adjoint=False):
        if adjoint is True:
            return (
                self._dhdtDeriv_u(tInd, src, v, adjoint),
                self._dhdtDeriv_m(tInd, src, v, adjoint)
            )
        return (
            self._dhdtDeriv_u(tInd, src, dun_dm_v) +
            self._dhdtDeriv_m(tInd, src, v)
        )

    def _jDeriv(self, tInd, src, dun_dm_v, v, adjoint=False):
        if adjoint is True:
            return (
                self._jDeriv_u(tInd, src, v, adjoint),
                self._jDeriv_m(tInd, src, v, adjoint)
            )
        return (
            self._jDeriv_u(tInd, src, dun_dm_v) +
            self._jDeriv_m(tInd, src, v)
        )


class Fields_Derivs_eb(FieldsTDEM):
    """
    A fields object for satshing derivs in the EB formulatio
    """
    knownFields = {
                    'bDeriv': 'F',
                    'eDeriv': 'E',
                    'hDeriv': 'F',
                    'jDeriv': 'E',
                    'dbdtDeriv': 'F',
                    'dhdtDeriv': 'F'
                  }


class Fields_Derivs_hj(FieldsTDEM):
    """
    A fields object for satshing derivs in the HJ formulation
    """
    knownFields = {
                    'bDeriv': 'E',
                    'eDeriv': 'F',
                    'hDeriv': 'E',
                    'jDeriv': 'F',
                    'dbdtDeriv': 'E',
                    'dhdtDeriv': 'E'
                  }


class Fields3D_b(FieldsTDEM):
    """Field Storage for a TDEM survey."""
    knownFields = {'bSolution': 'F'}
    aliasFields = {
                    'b': ['bSolution', 'F', '_b'],
                    'h': ['bSolution', 'F', '_h'],
                    'e': ['bSolution', 'E', '_e'],
                    'j': ['bSolution', 'E', '_j'],
                    'dbdt': ['bSolution', 'F', '_dbdt'],
                    'dhdt': ['bSolution', 'F', '_dhdt']
                  }

    def startup(self):
        self._MeSigma = self.survey.prob.MeSigma
        self._MeSigmaI = self.survey.prob.MeSigmaI
        self._MeSigmaDeriv = self.survey.prob.MeSigmaDeriv
        self._MeSigmaIDeriv = self.survey.prob.MeSigmaIDeriv
        self._edgeCurl = self.survey.prob.mesh.edgeCurl
        self._MfMui = self.survey.prob.MfMui
        self._timeMesh = self.survey.prob.timeMesh

    def _TLoc(self, fieldType):
        return 'N'

    def _b(self, bSolution, srcList, tInd):
        return bSolution

    def _bDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        return dun_dm_v

    def _bDeriv_m(self, tInd, src, v, adjoint=False):
        return Zero()

    def _dbdt(self, bSolution, srcList, tInd):
        # self._timeMesh.faceDiv
        dbdt = - self._edgeCurl * self._e(bSolution, srcList, tInd)
        for i, src in enumerate(srcList):
            s_m = src.s_m(self.survey.prob, self.survey.prob.times[tInd])
            dbdt[:, i] = dbdt[:, i] + s_m
        return dbdt

    def _dbdtDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint is True:
            return - self._eDeriv_u(
                tInd, src, self._edgeCurl.T * dun_dm_v, adjoint
            )
        return -(self._edgeCurl * self._eDeriv_u(tInd, src, dun_dm_v))

    def _dbdtDeriv_m(self, tInd, src, v, adjoint=False):
        if adjoint is True:
            return -(self._eDeriv_m(tInd, src, self._edgeCurl.T * v, adjoint))
        return -(self._edgeCurl * self._eDeriv_m(tInd, src, v)) #+ src.s_mDeriv() assuming src doesn't have deriv for now

    def _e(self, bSolution, srcList, tInd):
        e = self._MeSigmaI * (self._edgeCurl.T * (self._MfMui * bSolution))
        for i, src in enumerate(srcList):
            s_e = src.s_e(self.survey.prob, self.survey.prob.times[tInd])
            e[:, i] = e[:, i] - self._MeSigmaI * s_e
        return e

    def _eDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint is True:
            return (
                self._MfMui.T *
                (self._edgeCurl * (self._MeSigmaI.T * dun_dm_v))
            )
        return (
            self._MeSigmaI * (self._edgeCurl.T * (self._MfMui * dun_dm_v))
        )

    def _eDeriv_m(self, tInd, src, v, adjoint=False):
        _, s_e = src.eval(self.survey.prob, self.survey.prob.times[tInd])
        bSolution = self[[src], 'bSolution', tInd].flatten()

        _, s_eDeriv = src.evalDeriv(
            self.survey.prob.times[tInd], self, adjoint=adjoint
        )

        if adjoint is True:
            return (
                self._MeSigmaIDeriv(
                    -s_e + self._edgeCurl.T * (self._MfMui * bSolution), v,
                    adjoint
                ) -
                s_eDeriv(self._MeSigmaI.T * v)
            )

        return (
            self._MeSigmaIDeriv(
                -s_e + self._edgeCurl.T * (self._MfMui * bSolution),
                v, adjoint
            ) - self._MeSigmaI * s_eDeriv(v)
        )

    def _j(self, hSolution, srcList, tInd):
        return self.survey.prob.MeI * (
            self._MeSigma * self._e(hSolution, srcList, tInd)
        )

    def _jDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint:
            return self._eDeriv_u(
                tInd, src,
                self._MeSigma.T * (self.survey.prob.MeI.T * dun_dm_v),
                adjoint=True
            )
        return self.survey.prob.MeI * (
            self._MeSigma * self._eDeriv_u(tInd, src, dun_dm_v)
        )

    def _jDeriv_m(self, tInd, src, v, adjoint=False):
        e = self[src, 'e', tInd]
        if adjoint:
            w = self.survey.prob.MeI.T * v
            return (
                self._MeSigmaDeriv(e).T * w +
                self._eDeriv_m(tInd, src, self._MeSigma.T * w, adjoint=True)
            )
        return self.survey.prob.MeI * (
            self._MeSigmaDeriv(e) * v +
            self._MeSigma * self._eDeriv_m(tInd, src, v)
        )

    def _h(self, hSolution, srcList, tInd):
        return self.survey.prob.MfI * (
            self._MfMui * self._b(hSolution, srcList, tInd)
        )

    def _hDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint:
            return self._bDeriv_u(
                tInd, src, self._MfMui.T * (self.survey.prob.MfI.T * dun_dm_v),
                adjoint=True
            )
        return self.survey.prob.MfI * (
            self._MfMui * self._bDeriv_u(tInd, src, dun_dm_v)
        )

    def _hDeriv_m(self, tInd, src, v, adjoint=False):
        if adjoint:
            return self._bDeriv_m(
                tInd, src, self._MfMui.T * (self.survey.prob.MfI.T * v),
                adjoint=True
            )
        return self.survey.prob.MfI * (
            self._MfMui * self._bDeriv_m(tInd, src, v)
        )

    def _dhdt(self, hSolution, srcList, tInd):
        return self.survey.prob.MfI * (
            self._MfMui * self._dbdt(hSolution, srcList, tInd)
        )

    def _dhdtDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint:
            return self._dbdtDeriv_u(
                tInd, src, self._MfMui.T * (self.survey.prob.MfI.T * dun_dm_v),
                adjoint=True
            )
        return self.survey.prob.MfI * (
            self._MfMui * self._dbdtDeriv_u(tInd, src, dun_dm_v)
        )

    def _dhdtDeriv_m(self, tInd, src, v, adjoint=False):
        if adjoint:
            return self._dbdtDeriv_m(
                tInd, src, self._MfMui.T * (self.survey.prob.MfI.T * v),
                adjoint=True
            )
        return self.survey.prob.MfI * (
            self._MfMui * self._dbdtDeriv_m(tInd, src, v)
        )


class Fields3D_e(FieldsTDEM):
    """Fancy Field Storage for a TDEM survey."""
    knownFields = {'eSolution': 'E'}
    aliasFields = {
                    'e': ['eSolution', 'E', '_e'],
                    'j': ['eSolution', 'E', '_j'],
                    'b': ['eSolution', 'F', '_b'],
                    # 'h': ['eSolution', 'F', '_h'],
                    'dbdt': ['eSolution', 'F', '_dbdt'],
                    'dhdt': ['eSolution', 'F', '_dhdt'],
                  }

    def startup(self):
        self._MeSigma = self.survey.prob.MeSigma
        self._MeSigmaI = self.survey.prob.MeSigmaI
        self._MeSigmaDeriv = self.survey.prob.MeSigmaDeriv
        self._MeSigmaIDeriv = self.survey.prob.MeSigmaIDeriv
        self._edgeCurl = self.survey.prob.mesh.edgeCurl
        self._MfMui = self.survey.prob.MfMui
        self._times = self.survey.prob.times

    def _TLoc(self, fieldType):
        return 'N'

    def _e(self, eSolution, srcList, tInd):
        return eSolution

    def _eDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        return dun_dm_v

    def _eDeriv_m(self, tInd, src, v, adjoint=False):
        return Zero()

    def _dbdt(self, eSolution, srcList, tInd):
        s_m = np.zeros((self.mesh.nF, len(srcList)))
        for i, src in enumerate(srcList):
            s_m_src = src.s_m(
                self.survey.prob, self._times[tInd]
            )
            s_m[:, i] = s_m[:, i] + s_m_src
        return s_m - self._edgeCurl * eSolution

    def _dbdtDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint:
            return -self._edgeCurl.T * dun_dm_v
        return -self._edgeCurl * dun_dm_v

    def _dbdtDeriv_m(self, tInd, src, v, adjoint=False):
        # s_mDeriv = src.s_mDeriv(
        #     self._times[tInd], self, adjoint=adjoint
        # )
        return Utils.Zero()  # assumes source doesn't depend on model

    def _b(self, eSolution, srcList, tInd):
        """
        Integrate _db_dt using rectangles
        """
        raise NotImplementedError('To obtain b-fields, please use Problem3D_b')
        # dbdt = self._dbdt(eSolution, srcList, tInd)
        # dt = self.survey.prob.timeMesh.hx
        # # assume widths of "ghost cells" same on either end
        # dtn = np.hstack([dt[0], 0.5*(dt[1:] + dt[:-1]), dt[-1]])
        # return dtn[tInd] * dbdt
        # # raise NotImplementedError

    def _j(self, eSolution, srcList, tInd):
        return self.survey.prob.MeI * (
            self._MeSigma * self._e(eSolution, srcList, tInd)
        )

    def _jDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint:
            return self._eDeriv_u(
                tInd, src,
                self._MeSigma.T * (self.survey.prob.MeI.T * dun_dm_v),
                adjoint=True
            )
        return self.survey.prob.MeI * (
            self._MeSigma * self._eDeriv_u(tInd, src, dun_dm_v)
        )

    def _jDeriv_m(self, tInd, src, v, adjoint=False):
        e = self[src, 'e', tInd]
        if adjoint:
            w = self.survey.prob.MeI.T * v
            return (
                self._MeSigmaDeriv(e).T * w +
                self._eDeriv_m(tInd, src, self._MeSigma.T * w, adjoint=True)
            )
        return self.survey.prob.MeI * (
            self._MeSigmaDeriv(e) * v +
            self._MeSigma * self._eDeriv_m(tInd, src, v)
        )

    def _dhdt(self, eSolution, srcList, tInd):
        return self.survey.prob.MfI * (
            self._MfMui * self._dbdt(eSolution, srcList, tInd)
        )

    def _dhdtDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint:
            return self._dbdtDeriv_u(
                tInd, src, self._MfMui.T * (self.survey.prob.MfI.T * dun_dm_v),
                adjoint=True
            )
        return self.survey.prob.MfI * (
            self._MfMui * self._dbdtDeriv_u(tInd, src, dun_dm_v)
        )

    def _dhdtDeriv_m(self, tInd, src, v, adjoint=False):
        if adjoint:
            return self._dbdtDeriv_m(
                tInd, src, self._MfMui.T * (self.survey.prob.MfI.T * v)
            )
        return self.survey.prob.MfI * (
            self._MfMui * self._dbdtDeriv_m(tInd, src, v)
        )


class Fields3D_h(FieldsTDEM):
    """Fancy Field Storage for a TDEM survey."""
    knownFields = {'hSolution': 'E'}
    aliasFields = {
                    'h': ['hSolution', 'E', '_h'],
                    'b': ['hSolution', 'E', '_b'],
                    'dhdt': ['hSolution', 'E', '_dhdt'],
                    'dbdt': ['hSolution', 'E', '_dbdt'],
                    'j': ['hSolution', 'F', '_j'],
                    'e': ['hSolution', 'F', '_e'],
                    'charge': ['hSolution', 'CC', '_charge']
                  }

    def startup(self):
        self._edgeCurl = self.survey.prob.mesh.edgeCurl
        self._times = self.survey.prob.times
        self._MeMuI = self.survey.prob.MeMuI
        self._MeMu = self.survey.prob.MeMu
        self._MfRho = self.survey.prob.MfRho
        self._MfRhoDeriv = self.survey.prob.MfRhoDeriv

    def _TLoc(self, fieldType):
        # if fieldType in ['h', 'j']:
        return 'N'
        # else:
        #     raise NotImplementedError

    def _h(self, hSolution, srcList, tInd):
        return hSolution

    def _hDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        return dun_dm_v

    def _hDeriv_m(self, tInd, src, v, adjoint=False):
        return Zero()

    def _dhdt(self, hSolution, srcList, tInd):
        C = self._edgeCurl
        MeMuI = self._MeMuI
        MfRho = self._MfRho

        dhdt = - MeMuI * (C.T * (MfRho * (C * hSolution)))

        for i, src in enumerate(srcList):
            s_m, s_e = src.eval(self.survey.prob, self._times[tInd])
            dhdt[:, i] = MeMuI * (C.T * MfRho * s_e + s_m) +  dhdt[:, i]
        return dhdt

    def _dhdtDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        C = self._edgeCurl
        MeMuI = self._MeMuI
        MfRho = self._MfRho

        if adjoint:
            return - C.T * (MfRho.T * (C * (MeMuI * dun_dm_v)))
        return - MeMuI * (C.T * (MfRho * (C * dun_dm_v)))

    def _dhdtDeriv_m(self, tInd, src, v, adjoint=False):
        C = self._edgeCurl
        MeMuI = self._MeMuI
        MfRho = self._MfRho
        MfRhoDeriv = self._MfRhoDeriv

        hSolution = self[[src], 'hSolution', tInd].flatten()
        s_e = src.s_e(self.survey.prob, self._times[tInd])

        if adjoint:
            return - MfRhoDeriv(
                C * hSolution - s_e, (C * (MeMuI * v)), adjoint
            )
        return - MeMuI * (
            C.T * (MfRhoDeriv(C * hSolution - s_e, v, adjoint))
        )

    def _j(self, hSolution, srcList, tInd):
        s_e = np.zeros((self.mesh.nF, len(srcList)))
        for i, src in enumerate(srcList):
            s_e_src = src.s_e(
                self.survey.prob, self._times[tInd]
            )
            s_e[:, i] = s_e[:, i] + s_e_src

        return self._edgeCurl * hSolution - s_e

    def _jDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint:
            return self._edgeCurl.T * dun_dm_v
        return self._edgeCurl * dun_dm_v

    def _jDeriv_m(self, tInd, src, v, adjoint=False):
        return Zero() # assumes the source doesn't depend on the model

    def _b(self, hSolution, srcList, tInd):
        h = self._h(hSolution, srcList, tInd)
        return self.survey.prob.MeI * (self._MeMu * h)

    def _bDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint:
            return self._hDeriv_u(
                tInd, src, self._MeMu.T * (self.survey.prob.MeI.T * dun_dm_v),
                adjoint=adjoint
            )
        return self.survey.prob.MeI * (self._MeMu * self._hDeriv_u(tInd, src, dun_dm_v))

    def _bDeriv_m(self, tInd, src, v, adjoint=False):
        if adjoint:
            return self._hDeriv_m(
                tInd, src, self._MeMu.T * (self.survey.prob.MeI.T * v), adjoint=adjoint
            )
        return self.survey.prob.MeI * (self._MeMu * self._hDeriv_m(tInd, src, v))

    def _dbdt(self, hSolution, srcList, tInd):
        dhdt = self._dhdt(hSolution, srcList, tInd)
        return self.survey.prob.MeI * (self._MeMu * dhdt)

    def _dbdtDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint:
            return self._dhdtDeriv_u(
                tInd, src, self._MeMu.T * (self.survey.prob.MeI.T * dun_dm_v),
                adjoint=adjoint
            )
        return self.survey.prob.MeI * (self._MeMu * self._dhdtDeriv_u(tInd, src, dun_dm_v))

    def _dbdtDeriv_m(self, tInd, src, v, adjoint=False):
        if adjoint:
            return self._dhdtDeriv_m(
                tInd, src, self._MeMu.T * (self.survey.prob.MeI.T * v), adjoint=adjoint
            )
        return self.survey.prob.MeI * (self._MeMu * self._dhdtDeriv_m(tInd, src, v))

    def _e(self, hSolution, srcList, tInd):
        return self.survey.prob.MfI * (
            self._MfRho * self._j(hSolution, srcList, tInd)
        )

    def _eDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint:
            return self._jDeriv_u(
                tInd, src,
                self._MfRho.T * (self.survey.prob.MfI.T * dun_dm_v),
                adjoint=True
            )
        return self.survey.prob.MfI * (
            self._MfRho * self._jDeriv_u(tInd, src, dun_dm_v)
        )

    def _eDeriv_m(self, tInd, src, v, adjoint=False):
        j = Utils.mkvc(self[src, 'j', tInd])
        if adjoint is True:
            return (
                self._MfRhoDeriv(j, self.survey.prob.MfI.T * v, adjoint) +
                self._jDeriv_m(tInd, src, self._MfRho * v)
            )
        return self.survey.prob.MfI * (
            self._MfRhoDeriv(j, v) +
            self._MfRho * self._jDeriv_m(tInd, src, v)
        )

    def _charge(self, hSolution, srcList, tInd):
        vol = sdiag(self.survey.prob.mesh.vol)
        return epsilon_0*vol*(
            self.survey.prob.mesh.faceDiv*self._e(hSolution, srcList, tInd)
        )


class Fields3D_j(FieldsTDEM):
    """Fancy Field Storage for a TDEM survey."""
    knownFields = {'jSolution': 'F'}
    aliasFields = {
                    'dhdt': ['jSolution', 'E', '_dhdt'],
                    'dbdt': ['jSolution', 'E', '_dbdt'],
                    'j': ['jSolution', 'F', '_j'],
                    'e': ['jSolution', 'F', '_e'],
                    'charge': ['jSolution', 'CC', '_charge'],
                    'charge_density': ['jSolution', 'CC', '_charge_density'],
                  }

    def startup(self):
        self._edgeCurl = self.survey.prob.mesh.edgeCurl
        self._times = self.survey.prob.times
        self._MeMuI = self.survey.prob.MeMuI
        self._MfRho = self.survey.prob.MfRho
        self._MfRhoDeriv = self.survey.prob.MfRhoDeriv

    def _TLoc(self, fieldType):
        # if fieldType in ['h', 'j']:
        return 'N'

    def _j(self, jSolution, srcList, tInd):
        return jSolution

    def _jDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        return dun_dm_v

    def _jDeriv_m(self, tInd, src, v, adjoint=False):
        return Zero()

    def _h(self, jSolution, srcList, tInd):
        raise NotImplementedError('Please use Problem3D_h to get h-fields')

    def _dhdt(self, jSolution, srcList, tInd):
        C = self._edgeCurl
        MfRho = self._MfRho
        MeMuI = self._MeMuI

        dhdt = - MeMuI * (C.T * (MfRho * jSolution))
        for i, src in enumerate(srcList):
            s_m = src.s_m(self.survey.prob, self.survey.prob.times[tInd])
            dhdt[:, i] = MeMuI * s_m + dhdt[:, i]
        return dhdt

    def _dhdtDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        C = self._edgeCurl
        MfRho = self._MfRho
        MeMuI = self._MeMuI

        if adjoint is True:
            return -MfRho.T * (C * (MeMuI.T * dun_dm_v))
        return -MeMuI * (C.T * (MfRho * dun_dm_v))

    def _dhdtDeriv_m(self, tInd, src, v, adjoint=False):
        jSolution = self[[src], 'jSolution', tInd].flatten()
        C = self._edgeCurl
        MeMuI = self._MeMuI

        if adjoint is True:
            return -self._MfRhoDeriv(jSolution, C * (MeMuI * v), adjoint)
        return -MeMuI * (C.T * (self._MfRhoDeriv(jSolution, v)))

    def _e(self, jSolution, srcList, tInd):
        return self.survey.prob.MfI * (
            self._MfRho * self._j(jSolution, srcList, tInd)
        )

    def _eDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        if adjoint is True:
            return self._MfRho.T * (self.survey.prob.MfI.T * dun_dm_v)
        return self.survey.prob.MfI * (self._MfRho * dun_dm_v)

    def _eDeriv_m(self, tInd, src, v, adjoint=False):
        jSolution = Utils.mkvc(self[src, 'jSolution', tInd])
        if adjoint:
            return self._MfRhoDeriv(
                jSolution, self.survey.prob.MfI.T * v, adjoint
            )
        return self.survey.prob.MfI * self._MfRhoDeriv(jSolution, v)

    def _charge(self, jSolution, srcList, tInd):
        vol = sdiag(self.survey.prob.mesh.vol)
        return vol*self._charge_density(jSolution, srcList, tInd)

    def _charge_density(self, jSolution, srcList, tInd):
        return epsilon_0*(
            self.survey.prob.mesh.faceDiv*self._e(jSolution, srcList, tInd)
        )

    def _dbdt(self, jSolution, srcList, tInd):
        dhdt = Utils.mkvc(self._dhdt(jSolution, srcList, tInd))
        return self.survey.prob.MeI * (self.survey.prob.MeMu * dhdt)

    def _dbdtDeriv_u(self, tInd, src, dun_dm_v, adjoint=False):
        # dhdt = Utils.mkvc(self[src, 'dhdt', tInd])
        if adjoint:
            return self._dhdtDeriv_u(
                tInd, src,
                self.survey.prob.MeMu.T * (self.survey.prob.MeI.T * dun_dm_v),
                adjoint
            )
        return self.survey.prob.MeI * (
            self.survey.prob.MeMu * self._dhdtDeriv_u(tInd, src, dun_dm_v)
        )

    def _dbdtDeriv_m(self, tInd, src, v, adjoint=False):
        if adjoint:
            return self._dhdtDeriv_m(
                tInd, src,
                self.survey.prob.MeMu.T * (self.survey.prob.MeI.T * v),
                adjoint
            )
        return self.survey.prob.MeI * (
            self.survey.prob.MeMu * self._dhdtDeriv_m(tInd, src, v)
        )




import numpy as np
import SimPEG as simpeg




# campo primário para octree    <<<<<<<<<(mod. Diego em 14 dez 2019)
def homo1DModelSource(mesh, freq, sigma_1d):
    """
    Function that calculates and return background fields

    :param Simpeg mesh object mesh: Holds information on the discretization
    :param float freq: The frequency to solve at
    :param np.array sigma_1d: Background model of conductivity to base the calculations on, 1d model.
    :rtype: numpy.ndarray (mesh.nE, 2)
    :return: eBG_bp, E fields for the background model at both polarizations.

    """
    from . import get1DEfields
    # from . import getEHfields   # <<<<<<<<<(mod. Diego em 17 dez 2019)
    # Get a 1d solution for a halfspace background
    if mesh.dim == 1:
        mesh1d = mesh
    elif mesh.dim == 2:
        mesh1d = simpeg.Mesh.TensorMesh([mesh.hy], np.array([mesh.x0[1]]))
    elif mesh.dim == 3:
        mesh1d = simpeg.Mesh.TensorMesh([mesh.hz], np.array([mesh.x0[2]]))   


    # # Note: Everything is using e^iwt
    
    e0_1d = get1DEfields(mesh1d, sigma_1d, freq) #, mesh.vectorNz)
    
    
    
#    Ed, Eu, Hd, Hu = getEHfields(mesh1d,sigma_1d,freq,mesh1d.vectorNx) # <<<<<<<<<(mod. Diego em 17 dez 2019)
#    e0_1d = Ed + Eu

    # sourceAmp = 1.0
    # if sourceAmp is not None:
    #     e0_1d = ((e0_1d/e0_1d[-1])*sourceAmp)    
    
    
    E1dFieldDict = dict(zip(mesh.vectorNz, e0_1d))
    if mesh.dim == 1:
        eBG_px = simpeg.mkvc(e0_1d, 2)
        eBG_py = -simpeg.mkvc(e0_1d, 2) # added a minus to make the results in the correct quadrents.
    elif mesh.dim == 2:
        ex_px = np.zeros(mesh.vnEx, dtype=complex)
        ey_px = np.zeros((mesh.nEy, 1), dtype=complex)
        for i in np.arange(mesh.vnEx[0]):
            ex_px[i, :] = -e0_1d
        eBG_px = np.vstack((simpeg.Utils.mkvc(ex_px, 2), ey_px))
        # Setup y (north) polarization (_py)
        ex_py = np.zeros((mesh.nEx, 1), dtype='complex128')
        ey_py = np.zeros(mesh.vnEy, dtype='complex128')
        # Assign the source to ey_py
        for i in np.arange(mesh.vnEy[0]):
            ey_py[i, :] = e0_1d
        # ey_py[1:-1, 1:-1, 1:-1] = 0
        eBG_py = np.vstack((ex_py, simpeg.Utils.mkvc(ey_py, 2), ez_py))
    elif mesh.dim == 3:
        # Setup x (east) polarization (_x)        
        ex_px = -np.array([E1dFieldDict[i] for i in mesh.gridEx[:, 2]]).reshape(-1, 1)
        ey_px = np.zeros((mesh.nEy, 1), dtype=complex)
        ez_px = np.zeros((mesh.nEz, 1), dtype=complex)
        # Construct the full fields
        eBG_px = np.vstack((ex_px, ey_px, ez_px))
        # Setup y (north) polarization (_py)
        ex_py = np.zeros((mesh.nEx, 1), dtype=complex)
        ey_py = np.array([E1dFieldDict[i] for i in mesh.gridEy[:, 2]]).reshape(-1, 1)
        ez_py = np.zeros((mesh.nEz, 1), dtype=complex)
        # Construct the full fields
        eBG_py = np.vstack((ex_py,ey_py,ez_py))
    
    # Return the electric fields
    eBG_bp = np.hstack((eBG_px, eBG_py))  
    return eBG_bp



#def homo1DModelSource(mesh, freq, sigma_1d):
#    """
#        Function that calculates and return background fields
#
#        :param Simpeg mesh object mesh: Holds information on the discretization
#        :param float freq: The frequency to solve at
#        :param np.array sigma_1d: Background model of conductivity to base the calculations on, 1d model.
#        :rtype: numpy.ndarray (mesh.nE, 2)
#        :return: eBG_bp, E fields for the background model at both polarizations.
#
#    """
#    from . import get1DEfields
#    # Get a 1d solution for a halfspace background
#    if mesh.dim == 1:
#        mesh1d = mesh
#    elif mesh.dim == 2:
#        mesh1d = simpeg.Mesh.TensorMesh([mesh.hy], np.array([mesh.x0[1]]))
#    elif mesh.dim == 3:
#        mesh1d = simpeg.Mesh.TensorMesh([mesh.hz], np.array([mesh.x0[2]]))
#
#    # # Note: Everything is using e^iwt
#    e0_1d = get1DEfields(mesh1d, sigma_1d, freq)
#    if mesh.dim == 1:
#        eBG_px = simpeg.mkvc(e0_1d, 2)
#        eBG_py = -simpeg.mkvc(e0_1d, 2) # added a minus to make the results in the correct quadrents.
#    elif mesh.dim == 2:
#        ex_px = np.zeros(mesh.vnEx, dtype=complex)
#        ey_px = np.zeros((mesh.nEy, 1), dtype=complex)
#        for i in np.arange(mesh.vnEx[0]):
#            ex_px[i, :] = -e0_1d
#        eBG_px = np.vstack((simpeg.Utils.mkvc(ex_px, 2), ey_px))
#        # Setup y (north) polarization (_py)
#        ex_py = np.zeros((mesh.nEx, 1), dtype='complex128')
#        ey_py = np.zeros(mesh.vnEy, dtype='complex128')
#        # Assign the source to ey_py
#        for i in np.arange(mesh.vnEy[0]):
#            ey_py[i, :] = e0_1d
#        # ey_py[1:-1, 1:-1, 1:-1] = 0
#        eBG_py = np.vstack((ex_py, simpeg.Utils.mkvc(ey_py, 2), ez_py))
#    elif mesh.dim == 3:
#        # Setup x (east) polarization (_x)
#        ex_px = np.zeros(mesh.vnEx, dtype=complex)
#        ey_px = np.zeros((mesh.nEy, 1), dtype=complex)
#        ez_px = np.zeros((mesh.nEz, 1), dtype=complex)
#        # Assign the source to ex_x
#        for i in np.arange(mesh.vnEx[0]):
#            for j in np.arange(mesh.vnEx[1]):
#                ex_px[i, j, :] = -e0_1d
#        eBG_px = np.vstack((simpeg.Utils.mkvc(ex_px, 2), ey_px, ez_px))
#        # Setup y (north) polarization (_py)
#        ex_py = np.zeros((mesh.nEx, 1), dtype='complex128')
#        ey_py = np.zeros(mesh.vnEy, dtype='complex128')
#        ez_py = np.zeros((mesh.nEz, 1), dtype='complex128')
#        # Assign the source to ey_py
#        for i in np.arange(mesh.vnEy[0]):
#            for j in np.arange(mesh.vnEy[1]):
#                ey_py[i, j, :] = e0_1d
#        # ey_py[1:-1, 1:-1, 1:-1] = 0
#        eBG_py = np.vstack((ex_py, simpeg.Utils.mkvc(ey_py, 2), ez_py))
#
#    # Return the electric fields
#    eBG_bp = np.hstack((eBG_px, eBG_py))
#    return eBG_bp


def analytic1DModelSource(mesh, freq, sigma_1d):
    """
        Function that calculates and return background fields

        :param Simpeg mesh object mesh: Holds information on the discretization
        :param float freq: The frequency to solve at
        :param np.array sigma_1d: Background model of conductivity to base the calculations on, 1d model.
        :rtype: numpy.ndarray (mesh.nE, 2)
        :return: eBG_bp, E fields for the background model at both polarizations.

    """
    from . import getEHfields
    # Get a 1d solution for a halfspace background
    if mesh.dim == 1:
        mesh1d = mesh
    elif mesh.dim == 2:
        mesh1d = simpeg.Mesh.TensorMesh([mesh.hy], np.array([mesh.x0[1]]))
    elif mesh.dim == 3:
        mesh1d = simpeg.Mesh.TensorMesh([mesh.hz], np.array([mesh.x0[2]]))

    # # Note: Everything is using e^iwt
    Eu, Ed, _, _ = getEHfields(mesh1d, sigma_1d, freq, mesh.vectorNz)
    # Make the fields into a dictionary of location and the fields
    e0_1d = Eu+Ed
    E1dFieldDict = dict(zip(mesh.vectorNz, e0_1d))
    if mesh.dim == 1:
        eBG_px = simpeg.mkvc(e0_1d, 2)
        eBG_py = -simpeg.mkvc(e0_1d, 2) # added a minus to make the results in the correct quadrents.
    elif mesh.dim == 2:
        ex_px = np.zeros(mesh.vnEx, dtype=complex)
        ey_px = np.zeros((mesh.nEy, 1), dtype=complex)
        for i in np.arange(mesh.vnEx[0]):
            ex_px[i, :] = -e0_1d
        eBG_px = np.vstack((simpeg.Utils.mkvc(ex_px, 2), ey_px))
        # Setup y (north) polarization (_py)
        ex_py = np.zeros((mesh.nEx, 1), dtype='complex128')
        ey_py = np.zeros(mesh.vnEy, dtype='complex128')
        # Assign the source to ey_py
        for i in np.arange(mesh.vnEy[0]):
            ey_py[i, :] = e0_1d
        # ey_py[1:-1, 1:-1, 1:-1] = 0
        eBG_py = np.vstack((ex_py, simpeg.Utils.mkvc(ey_py, 2), ez_py))
    elif mesh.dim == 3:
        # Setup x (east) polarization (_x)
        ex_px = -np.array([E1dFieldDict[i] for i in mesh.gridEx[:, 2]]).reshape(-1, 1)
        ey_px = np.zeros((mesh.nEy, 1), dtype=complex)
        ez_px = np.zeros((mesh.nEz, 1), dtype=complex)
        # Construct the full fields
        eBG_px = np.vstack((ex_px, ey_px, ez_px))
        # Setup y (north) polarization (_py)
        ex_py = np.zeros((mesh.nEx, 1), dtype='complex128')
        ey_py = np.array([E1dFieldDict[i] for i in mesh.gridEy[:, 2]]).reshape(-1, 1)
        ez_py = np.zeros((mesh.nEz, 1), dtype='complex128')
        # Construct the full fields
        eBG_py = np.vstack((ex_py, simpeg.Utils.mkvc(ey_py, 2), ez_py))

    # Return the electric fields
    eBG_bp = np.hstack((eBG_px, eBG_py))
    return eBG_bp

def homo3DModelSource(mesh, model, freq):
     """
         Function that estimates 1D analytic background fields from a 3D model.

         :param Simpeg mesh object mesh: Holds information on the discretization
         :param float freq: The frequency to solve at
         :param np.array sigma_1d: Background model of conductivity to base the calculations on, 1d model.
         :rtype: numpy.ndarray (mesh.nE, 2)
         :return: eBG_bp, E fields for the background model at both polarizations.

     """

     if mesh.dim < 3:
         raise IOError('Input mesh has to have 3 dimensions.')


     # Get the locations
     a = mesh.gridCC[:, 0:2].copy()
     unixy = np.unique(a.view(a.dtype.descr * a.shape[1])).view(float).reshape(-1, 2)
     uniz = np.unique(mesh.gridCC[:, 2])
     # # Note: Everything is using e^iwt
     # Need to loop thourgh the xy locations, assess the model and calculate the fields at the phusdo cell centers.
     # Then interpolate the cc fields to the edges.

     e0_1d = get1DEfields(mesh1d, sigma_1d, freq)

#     elif mesh.dim == 3:
     # Setup x (east) polarization (_x)
     ex_px = np.zeros(mesh.vnEx, dtype=complex)
     ey_px = np.zeros((mesh.nEy, 1), dtype=complex)
     ez_px = np.zeros((mesh.nEz, 1), dtype=complex)
     # Assign the source to ex_x
     for i in np.arange(mesh.vnEx[0]):
         for j in np.arange(mesh.vnEx[1]):
             ex_px[i, j, :] = -e0_1d
     eBG_px = np.vstack((simpeg.Utils.mkvc(ex_px, 2), ey_px, ez_px))
     # Setup y (north) polarization (_py)
     ex_py = np.zeros((mesh.nEx, 1), dtype='complex128')
     ey_py = np.zeros(mesh.vnEy, dtype='complex128')
     ez_py = np.zeros((mesh.nEz, 1), dtype='complex128')
     # Assign the source to ey_py
     for i in np.arange(mesh.vnEy[0]):
         for j in np.arange(mesh.vnEy[1]):
             ey_py[i, j, :] = e0_1d
     # ey_py[1:-1, 1:-1, 1:-1] = 0
     eBG_py = np.vstack((ex_py, simpeg.Utils.mkvc(ey_py, 2), ez_py))

     # Return the electric fields
     eBG_bp = np.hstack((eBG_px, eBG_py))
     return eBG_bp



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from SimPEG import Utils
from SimPEG.EM.Base import BaseEMProblem
from .SurveyDC import Survey_ky
from .FieldsDC_2D import Fields_ky, Fields_ky_CC, Fields_ky_N
from .FieldsDC import FieldsDC, Fields_CC, Fields_N
import numpy as np
from SimPEG.Utils import Zero
from .BoundaryUtils import getxBCyBC_CC
from scipy.special import kn


class BaseDCProblem_2D(BaseEMProblem):
    """
    Base 2.5D DC problem
    """
    surveyPair = Survey_ky
    fieldsPair = Fields_ky  # SimPEG.EM.Static.Fields_2D
    fieldsPair_fwd = FieldsDC
    nky = 15
    kys = np.logspace(-4, 1, nky)
    Ainv = [None for i in range(nky)]
    nT = nky  # Only for using TimeFields
    storeJ = False
    _Jmatrix = None
    fix_Jmatrix = False

    def fields(self, m):
        if self.verbose:
            print (">> Compute fields")
        if m is not None:
            self.model = m
        if self.Ainv[0] is not None:
            for i in range(self.nky):
                self.Ainv[i].clean()
        f = self.fieldsPair(self.mesh, self.survey)
        Srcs = self.survey.srcList
        for iky in range(self.nky):
            ky = self.kys[iky]
            A = self.getA(ky)
            self.Ainv[iky] = self.Solver(A, **self.solverOpts)
            RHS = self.getRHS(ky)
            u = self.Ainv[iky] * RHS
            f[Srcs, self._solutionType, iky] = u
        return f

    def fields_to_space(self, f, y=0.):
        f_fwd = self.fieldsPair_fwd(self.mesh, self.survey)
        # Evaluating Integration using Trapezoidal rules
        nky = self.kys.size
        dky = np.diff(self.kys)
        dky = np.r_[dky[0], dky]
        phi0 = 1./np.pi*f[:, self._solutionType, 0]
        phi = np.zeros_like(phi0)
        for iky in range(nky):
            phi1 = 1./np.pi*f[:, self._solutionType, iky]
            phi += phi1*dky[iky]/2.*np.cos(self.kys[iky]*y)
            phi += phi0*dky[iky]/2.*np.cos(self.kys[iky]*y)
            phi0 = phi1.copy()
        f_fwd[:, self._solutionType] = phi
        return f_fwd

    def getJ(self, m, f=None):
        """
            Generate Full sensitivity matrix
        """
        if self._Jmatrix is not None:
            return self._Jmatrix
        else:
            if self.verbose:
                print("Calculating J and storing")
            self.model = m
            if f is None:
                f = self.fields(m)
            self._Jmatrix = (self._Jtvec(m, v=None, f=f)).T
        return self._Jmatrix

    def Jvec(self, m, v, f=None):
        """
            Compute sensitivity matrix (J) and vector (v) product.
        """
        if self.storeJ:
            J = self.getJ(m, f=f)
            Jv = Utils.mkvc(np.dot(J, v))
            return Jv

        self.model = m

        if f is None:
            f = self.fields(m)

        # TODO: This is not a good idea !! should change that as a list
        Jv = self.dataPair(self.survey)  # same size as the data
        Jv0 = self.dataPair(self.survey)

        # Assume y=0.
        # This needs some thoughts to implement in general when src is dipole
        dky = np.diff(self.kys)
        dky = np.r_[dky[0], dky]
        y = 0.

        # TODO: this loop is pretty slow .. (Parellize)
        for iky in range(self.nky):
            ky = self.kys[iky]
            for src in self.survey.srcList:
                u_src = f[src, self._solutionType, iky]  # solution vector
                dA_dm_v = self.getADeriv(ky, u_src, v, adjoint=False)
                dRHS_dm_v = self.getRHSDeriv(ky, src, v)
                du_dm_v = self.Ainv[iky] * (- dA_dm_v + dRHS_dm_v)
                for rx in src.rxList:
                    df_dmFun = getattr(f, '_{0!s}Deriv'.format(rx.projField),
                                       None)
                    df_dm_v = df_dmFun(iky, src, du_dm_v, v, adjoint=False)
                    # Trapezoidal intergration
                    Jv1_temp = 1./np.pi*rx.evalDeriv(ky, src, self.mesh, f,
                                                     df_dm_v)
                    if iky == 0:
                        # First assigment
                        Jv[src, rx] = Jv1_temp*dky[iky]*np.cos(ky*y)
                    else:
                        Jv[src, rx] += Jv1_temp*dky[iky]/2.*np.cos(ky*y)
                        Jv[src, rx] += Jv0[src, rx]*dky[iky]/2.*np.cos(ky*y)
                    Jv0[src, rx] = Jv1_temp.copy()
        return Utils.mkvc(Jv)

    def Jtvec(self, m, v, f=None):
        """
            Compute adjoint sensitivity matrix (J^T) and vector (v) product.
        """
        if self.storeJ:
            J = self.getJ(m, f=f)
            Jtv = Utils.mkvc(np.dot(J.T, v))
            return Jtv

        self.model = m

        if f is None:
            f = self.fields(m)

        return self._Jtvec(m, v=v, f=f)

    def _Jtvec(self, m, v=None, f=None):
        """
            Compute adjoint sensitivity matrix (J^T) and vector (v) product.
            Full J matrix can be computed by inputing v=None
        """

        if v is not None:
            # Ensure v is a data object.
            if not isinstance(v, self.dataPair):
                v = self.dataPair(self.survey, v)
            Jtv = np.zeros(m.size, dtype=float)

            # Assume y=0.
            dky = np.diff(self.kys)
            dky = np.r_[dky[0], dky]
            y = 0.

            for src in self.survey.srcList:
                for rx in src.rxList:
                    Jtv_temp1 = np.zeros(m.size, dtype=float)
                    Jtv_temp0 = np.zeros(m.size, dtype=float)

                    # TODO: this loop is pretty slow .. (Parellize)
                    for iky in range(self.nky):
                        u_src = f[src, self._solutionType, iky]
                        ky = self.kys[iky]
                        # wrt f, need possibility wrt m
                        PTv = rx.evalDeriv(ky, src, self.mesh, f, v[src, rx],
                                           adjoint=True)
                        df_duTFun = getattr(
                            f, '_{0!s}Deriv'.format(rx.projField), None
                        )
                        df_duT, df_dmT = df_duTFun(iky, src, None, PTv,
                                                   adjoint=True)

                        ATinvdf_duT = self.Ainv[iky] * df_duT

                        dA_dmT = self.getADeriv(ky, u_src, ATinvdf_duT,
                                                adjoint=True)
                        dRHS_dmT = self.getRHSDeriv(ky, src, ATinvdf_duT,
                                                    adjoint=True)
                        du_dmT = -dA_dmT + dRHS_dmT
                        Jtv_temp1 = 1./np.pi*(df_dmT + du_dmT).astype(float)
                        # Trapezoidal intergration
                        if iky == 0:
                            # First assigment
                            Jtv += Jtv_temp1*dky[iky]*np.cos(ky*y)
                        else:
                            Jtv += Jtv_temp1*dky[iky]/2.*np.cos(ky*y)
                            Jtv += Jtv_temp0*dky[iky]/2.*np.cos(ky*y)
                        Jtv_temp0 = Jtv_temp1.copy()
            return Utils.mkvc(Jtv)

        # This is for forming full sensitivity
        else:

            # This is for forming full sensitivity matrix
            Jt = np.zeros((self.model.size, self.survey.nD), order='F')
            istrt = int(0)
            iend = int(0)

            # Assume y=0.
            dky = np.diff(self.kys)
            dky = np.r_[dky[0], dky]
            y = 0.
            for src in self.survey.srcList:
                for rx in src.rxList:
                    iend = istrt + rx.nD
                    Jtv_temp1 = np.zeros((m.size, rx.nD), dtype=float)
                    Jtv_temp0 = np.zeros((m.size, rx.nD), dtype=float)
                    # TODO: this loop is pretty slow .. (Parellize)
                    for iky in range(self.nky):
                        u_src = f[src, self._solutionType, iky]
                        ky = self.kys[iky]
                        # wrt f, need possibility wrt m
                        P = rx.getP(self.mesh, rx.projGLoc(f)).toarray()

                        ATinvdf_duT = self.Ainv[iky] * (P.T)

                        dA_dmT = self.getADeriv(ky, u_src, ATinvdf_duT,
                                                adjoint=True)
                        Jtv_temp1 = 1./np.pi*(-dA_dmT)
                        # Trapezoidal intergration
                        if iky == 0:
                            # First assigment
                            if rx.nD == 1:
                                Jt[:, istrt] += Jtv_temp1*dky[iky]*np.cos(ky*y)
                            else:
                                Jt[:, istrt:iend] += Jtv_temp1*dky[iky]*np.cos(ky*y)
                        else:
                            if rx.nD == 1:
                                Jt[:, istrt] += Jtv_temp1*dky[iky]/2.*np.cos(ky*y)
                                Jt[:, istrt] += Jtv_temp0*dky[iky]/2.*np.cos(ky*y)
                            else:
                                Jt[:, istrt:iend] += Jtv_temp1*dky[iky]/2.*np.cos(ky*y)
                                Jt[:, istrt:iend] += Jtv_temp0*dky[iky]/2.*np.cos(ky*y)
                        Jtv_temp0 = Jtv_temp1.copy()
                    istrt += rx.nD
            return Jt

    def getSourceTerm(self, ky):
        """
        takes concept of source and turns it into a matrix
        """
        """
        Evaluates the sources, and puts them in matrix form
        :rtype: (numpy.ndarray, numpy.ndarray)
        :return: q (nC or nN, nSrc)
        """

        Srcs = self.survey.srcList

        if self._formulation == 'EB':
            n = self.mesh.nN
            # return NotImplementedError

        elif self._formulation == 'HJ':
            n = self.mesh.nC

        q = np.zeros((n, len(Srcs)))

        for i, src in enumerate(Srcs):
            q[:, i] = src.eval(self)
        return q

    @property
    def deleteTheseOnModelUpdate(self):
        toDelete = super(BaseDCProblem_2D, self).deleteTheseOnModelUpdate
        if self.sigmaMap is not None:
            toDelete += [
                '_MnSigma', '_MnSigmaDerivMat',
                '_MccRhoi', '_MccRhoiDerivMat'
            ]

        if self.fix_Jmatrix:
            return toDelete

        if self._Jmatrix is not None:
            toDelete += ['_Jmatrix']
        return toDelete

    ####################################################
    # Mass Matrices
    ####################################################

    @property
    def MnSigma(self):
        """
            Node inner product matrix for \\(\\sigma\\). Used in the E-B
            formulation
        """
        # TODO: only works isotropic sigma
        if getattr(self, '_MnSigma', None) is None:
            sigma = self.sigma
            vol = self.mesh.vol
            self._MnSigma = Utils.sdiag(
                self.mesh.aveN2CC.T*(Utils.sdiag(vol)*sigma)
            )
        return self._MnSigma

    @property
    def MnSigmaDerivMat(self):
        """
            Derivative of MnSigma with respect to the model
        """
        if getattr(self, '_MnSigmaDerivMat', None) is None:
            sigma = self.sigma
            vol = self.mesh.vol
            self._MnSigmaDerivMat = (
                self.mesh.aveN2CC.T * Utils.sdiag(vol) * self.sigmaDeriv
                )
        return self._MnSigmaDerivMat

    def MnSigmaDeriv(self, u, v, adjoint=False):
        """
            Derivative of MnSigma with respect to the model times a vector (u)
        """
        if self.storeInnerProduct:
            if adjoint:
                return self.MnSigmaDerivMat.T * (
                    Utils.sdiag(u)*v
                )
            else:
                return u*(self.MnSigmaDerivMat * v)
        else:
            sigma = self.sigma
            vol = self.mesh.vol
            if adjoint:
                return self.sigmaDeriv.T * (
                    Utils.sdiag(vol) * (self.mesh.aveN2CC * (Utils.sdiag(u)*v))
                )
            else:
                dsig_dm_v = self.sigmaDeriv * v
                return (
                    u * (self.mesh.aveN2CC.T * (vol * dsig_dm_v))
                )

    @property
    def MccRhoi(self):
        """
            Cell inner product matrix for \\(\\rho^{-1}\\). Used in the H-J
            formulation
        """
        # TODO: only works isotropic rho
        if getattr(self, '_MccRhoi', None) is None:
            self._MccRhoi = Utils.sdiag(
                self.mesh.vol/self.rho
            )
        return self._MccRhoi

    @property
    def MccRhoiDerivMat(self):
        """
            Derivative of MccRho with respect to the model
        """
        if getattr(self, '_MccRhoiDerivMat', None) is None:
            rho = self.rho
            vol = self.mesh.vol
            self._MccRhoiDerivMat = (
                Utils.sdiag(vol*(-1./rho**2))*self.rhoDeriv
            )
        return self._MccRhoiDerivMat

    def MccRhoiDeriv(self, u, v, adjoint=False):
        """
            Derivative of :code:`MccRhoi` with respect to the model.
        """
        if self.rhoMap is None:
            return Utils.Zero()

        if len(self.rho.shape) > 1:
            if self.rho.shape[1] > self.mesh.dim:
                raise NotImplementedError(
                    "Full anisotropy is not implemented for MccRhoiDeriv."
                )
        if self.storeInnerProduct:
            if adjoint:
                return self.MccRhoiDerivMat.T * (Utils.sdiag(u) * v)
            else:
                return Utils.sdiag(u) * (self.MccRhoiDerivMat * v)
        else:
            vol = self.mesh.vol
            rho = self.rho
            if adjoint:
                return self.rhoDeriv.T * (Utils.sdiag(u*vol*(-1./rho**2)) * v)
            else:
                return (Utils.sdiag(u*vol*(-1./rho**2)))*(self.rhoDeriv * v)


class Problem2D_CC(BaseDCProblem_2D):
    """
    2.5D cell centered DC problem
    """

    _solutionType = 'phiSolution'
    _formulation = 'HJ'  # CC potentials means J is on faces
    fieldsPair = Fields_ky_CC
    fieldsPair_fwd = Fields_CC
    bc_type = 'Mixed'

    def __init__(self, mesh, **kwargs):
        BaseDCProblem_2D.__init__(self, mesh, **kwargs)

    def getA(self, ky):
        """
        Make the A matrix for the cell centered DC resistivity problem
        A = D MfRhoI G
        """
        # To handle Mixed boundary condition
        self.setBC(ky=ky)
        D = self.Div
        G = self.Grad
        vol = self.mesh.vol
        MfRhoI = self.MfRhoI
        # Get resistivity rho
        rho = self.rho
        A = D * MfRhoI * G + ky**2 * self.MccRhoi
        if self.bc_type == "Neumann":
            A[0, 0] = A[0, 0] + 1.
        return A

    def getADeriv(self, ky, u, v, adjoint=False):
        # To handle Mixed boundary condition
        # self.setBC(ky=ky)

        D = self.Div
        G = self.Grad
        vol = self.mesh.vol
        if adjoint:
            return (
                self.MfRhoIDeriv(G*u.flatten(), D.T*v, adjoint=adjoint) +
                ky**2 * self.MccRhoiDeriv(u.flatten(), v, adjoint=adjoint)
            )
        else:
            return (
                D * self.MfRhoIDeriv(G*u.flatten(), v, adjoint=adjoint) +
                ky**2 * self.MccRhoiDeriv(u.flatten(), v, adjoint=adjoint)
            )

    def getRHS(self, ky):
        """
        RHS for the DC problem
        q
        """

        RHS = self.getSourceTerm(ky)
        return RHS

    def getRHSDeriv(self, ky, src, v, adjoint=False):
        """
        Derivative of the right hand side with respect to the model
        """
        # TODO: add qDeriv for RHS depending on m
        # qDeriv = src.evalDeriv(self, ky, adjoint=adjoint)
        # return qDeriv
        return Zero()

    def setBC(self, ky=None):
        fxm, fxp, fym, fyp = self.mesh.faceBoundaryInd
        gBFxm = self.mesh.gridFx[fxm, :]
        gBFxp = self.mesh.gridFx[fxp, :]
        gBFym = self.mesh.gridFy[fym, :]
        gBFyp = self.mesh.gridFy[fyp, :]

        # Setup Mixed B.C (alpha, beta, gamma)
        temp_xm = np.ones_like(gBFxm[:, 0])
        temp_xp = np.ones_like(gBFxp[:, 0])
        temp_ym = np.ones_like(gBFym[:, 1])
        temp_yp = np.ones_like(gBFyp[:, 1])

        if self.bc_type == "Neumann":
            alpha_xm, alpha_xp = temp_xm*0., temp_xp*0.
            alpha_ym, alpha_yp = temp_ym*0., temp_yp*0.

            beta_xm, beta_xp = temp_xm, temp_xp
            beta_ym, beta_yp = temp_ym, temp_yp

            gamma_xm, gamma_xp = temp_xm*0., temp_xp*0.
            gamma_ym, gamma_yp = temp_ym*0., temp_yp*0.

        elif self.bc_type == "Dirichlet":
            alpha_xm, alpha_xp = temp_xm, temp_xp
            alpha_ym, alpha_yp = temp_ym, temp_yp

            beta_xm, beta_xp = temp_xm*0., temp_xp*0.
            beta_ym, beta_yp = temp_ym*0., temp_yp*0.

            gamma_xm, gamma_xp = temp_xm*0., temp_xp*0.
            gamma_ym, gamma_yp = temp_ym*0., temp_yp*0.

        elif self.bc_type == "Mixed":
            xs = np.median(self.mesh.vectorCCx)
            ys = np.median(self.mesh.vectorCCy[-1])

            def r_boundary(x, y):
                return 1./np.sqrt(
                    (x - xs)**2 + (y - ys)**2
                    )

            rxm = r_boundary(gBFxm[:, 0], gBFxm[:, 1])
            rxp = r_boundary(gBFxp[:, 0], gBFxp[:, 1])
            rym = r_boundary(gBFym[:, 0], gBFym[:, 1])

            alpha_xm = ky*(
                kn(1, ky*rxm) / kn(0, ky*rxm) * (gBFxm[:, 0]-xs)
                )
            alpha_xp = ky*(
                kn(1, ky*rxp) / kn(0, ky*rxp) * (gBFxp[:, 0]-xs)
                )
            alpha_ym = ky*(
                kn(1, ky*rym) / kn(0, ky*rym) * (gBFym[:, 0]-ys)
                )
            alpha_yp = temp_yp*0.
            beta_xm, beta_xp = temp_xm, temp_xp
            beta_ym, beta_yp = temp_ym, temp_yp

            gamma_xm, gamma_xp = temp_xm*0., temp_xp*0.
            gamma_ym, gamma_yp = temp_ym*0., temp_yp*0.

        alpha = [alpha_xm, alpha_xp, alpha_ym, alpha_yp]
        beta = [beta_xm, beta_xp, beta_ym, beta_yp]
        gamma = [gamma_xm, gamma_xp, gamma_ym, gamma_yp]

        x_BC, y_BC = getxBCyBC_CC(self.mesh, alpha, beta, gamma)
        V = self.Vol
        self.Div = V * self.mesh.faceDiv
        P_BC, B = self.mesh.getBCProjWF_simple()
        M = B*self.mesh.aveCC2F
        self.Grad = self.Div.T - P_BC*Utils.sdiag(y_BC)*M


class Problem2D_N(BaseDCProblem_2D):
    """
    2.5D nodal DC problem
    """

    _solutionType = 'phiSolution'
    _formulation = 'EB'  # CC potentials means J is on faces
    fieldsPair = Fields_ky_N
    fieldsPair_fwd = Fields_N

    def __init__(self, mesh, **kwargs):
        BaseDCProblem_2D.__init__(self, mesh, **kwargs)
        # self.setBC()

    def getA(self, ky):
        """
        Make the A matrix for the cell centered DC resistivity problem
        A = D MfRhoI G
        """
        MeSigma = self.MeSigma
        MnSigma = self.MnSigma
        Grad = self.mesh.nodalGrad
        # Get conductivity sigma
        sigma = self.sigma
        A = Grad.T * MeSigma * Grad + ky**2*MnSigma
        # This seems not required for 2.5D problem
        # Handling Null space of A
        # A[0, 0] = A[0, 0] + 1.
        return A

    def getADeriv(self, ky, u, v, adjoint=False):

        MeSigma = self.MeSigma
        Grad = self.mesh.nodalGrad
        sigma = self.sigma
        vol = self.mesh.vol

        if adjoint:
            return (
                self.MeSigmaDeriv(Grad*u.flatten(), Grad*v, adjoint=adjoint) +
                ky**2*self.MnSigmaDeriv(u.flatten(), v, adjoint=adjoint)
            )
        else:
            return (
                Grad.T*self.MeSigmaDeriv(Grad*u.flatten(), v, adjoint=adjoint) +
                ky**2*self.MnSigmaDeriv(u.flatten(), v, adjoint=adjoint)
            )
        # return (Grad.T*(self.MeSigmaDeriv(Grad*u.flatten(), v, adjoint)) +
        #         ky**2*self.MnSigmaDeriv(u.flatten())*v)

    def getRHS(self, ky):
        """
        RHS for the DC problem
        q
        """

        RHS = self.getSourceTerm(ky)
        return RHS

    def getRHSDeriv(self, ky, src, v, adjoint=False):
        """
        Derivative of the right hand side with respect to the model
        """
        # TODO: add qDeriv for RHS depending on m
        # qDeriv = src.evalDeriv(self, ky, adjoint=adjoint)
        # return qDeriv
        return Zero()



from SimPEG.EM.Static.DC import Src
from SimPEG import Props
from SimPEG.Utils import sdiag
from SimPEG import Utils
import scipy.sparse as sp
import numpy as np
from SimPEG.EM.Static.DC import Survey


class StreamingCurrents(Src.BaseSrc):

    L = None
    mesh = None
    modelType = None
    indActive = None

    def __init__(self, rxList, **kwargs):
        Src.BaseSrc.__init__(self, rxList, **kwargs)
        if self.modelType == "Head":
            if self.L is None:
                raise Exception(
                    "SP source requires cross coupling coefficient L"
                )
        elif self.modelType == "CurrentDensity":
            if self.indActive is None:
                self.indActive = np.ones(self.mesh.nC, dtype=bool)
            # This is for setting a Neuman condition on the topographic faces

            self.Grad = -sp.vstack(
                (
                    self.Pafx * self.mesh.faceDivx.T * self.V * self.Pac,
                    self.Pafy * self.mesh.faceDivy.T * self.V * self.Pac,
                    self.Pafz * self.mesh.faceDivz.T * self.V * self.Pac,
                )
            )

        if self.mesh is None:
            raise Exception("SP source requires mesh")

    def getq_from_j(self, j):
        q = self.Grad.T*self.mesh.aveCCV2F*j
        return q

    def eval(self, prob):
        """

            Computing source term using:

            - Hydraulic head: h
            - Cross coupling coefficient: L

            .. math::

                -\nabla \cdot \vec{j}^s = \nabla \cdot L \nabla \phi \\

        """
        if prob._formulation == 'HJ':
            if self.modelType == "Head":
                q = prob.Grad.T*self.MfLiI*prob.Grad*prob.h
            elif self.modelType == "CurrentSource":
                q = self.V * prob.q
            elif self.modelType == "CurrentDensity":
                q = self.Grad.T*self.mesh.aveCCV2F*np.r_[
                    prob.jsx, prob.jsy, prob.jsz
                ]
            else:
                raise NotImplementedError()
        elif prob._formulation == 'EB':
            raise NotImplementedError()
        return q

    def evalDeriv(self, prob, v=None, adjoint=False):
        if prob._formulation == 'HJ':
            if adjoint:
                if self.modelType == "Head":
                    srcDeriv = prob.hDeriv.T * prob.Grad.T * self.MfLiI.T * (
                        prob.Grad * v
                    )
                elif self.modelType == "CurrentSource":
                    srcDeriv = prob.qDeriv.T * (self.V * v)
                elif self.modelType == "CurrentDensity":
                    jsDeriv = sp.vstack(
                        (prob.jsxDeriv, prob.jsyDeriv, prob.jszDeriv)
                    )
                    srcDeriv = jsDeriv.T * self.mesh.aveCCV2F.T * (self.Grad*v)
                else:
                    raise NotImplementedError()
            else:
                if self.modelType == "Head":
                    srcDeriv = prob.Grad.T*self.MfLiI*prob.Grad*(prob.hDeriv*v)
                elif self.modelType == "CurrentSource":
                    srcDeriv = self.V * (prob.qDeriv * v)
                elif self.modelType == "CurrentDensity":
                    jsDeriv = sp.vstack(
                        (prob.jsxDeriv, prob.jsyDeriv, prob.jszDeriv)
                    )
                    srcDeriv = self.Grad.T * self.mesh.aveCCV2F*(jsDeriv*v)
                else:
                    raise NotImplementedError()
        elif prob._formulation == 'EB':
            raise NotImplementedError()
        return srcDeriv
    @property
    def V(self):
        """
            :code:`V`
        """
        if getattr(self, '_V', None) is None:
            self._V = Utils.sdiag(self.mesh.vol)
        return self._V

    @property
    def MfLi(self):
        """
            :code:`MfLi`
        """
        if getattr(self, '_MfLi', None) is None:
            self._MfLi = self.mesh.getFaceInnerProduct(1./self.L)
        return seself.lf._MfLi

    @property
    def MfLiI(self):
        """
            Inverse of :code:`_MfLiI`
        """
        if getattr(self, '_MfLiI', None) is None:
            self._MfLiI = self.mesh.getFaceInnerProduct(1./self.L, invMat=True)
        return self._MfLiI

    @property
    def Pac(self):
        """
        diagonal matrix that nulls out inactive cells

        :rtype: scipy.sparse.csr_matrix
        :return: active cell diagonal matrix
        """
        if getattr(self, '_Pac', None) is None:
            if self.indActive is None:
                self._Pac = Utils.speye(self.mesh.nC)
            else:
                e = np.zeros(self.mesh.nC)
                e[self.indActive] = 1.
                # self._Pac = Utils.speye(self.mesh.nC)[:, self.indActive]
                self._Pac = Utils.sdiag(e)
        return self._Pac

    @property
    def Pafx(self):
        """
        diagonal matrix that nulls out inactive x-faces
        to full modelling space (ie. nFx x nindActive_Fx )

        :rtype: scipy.sparse.csr_matrix
        :return: active face-x diagonal matrix
        """
        if getattr(self, '_Pafx', None) is None:
            if self.indActive is None:
                self._Pafx = Utils.speye(self.mesh.nFx)
            else:
                indActive_Fx = self.mesh.aveFx2CC.T * self.indActive >= 1
                e = np.zeros(self.mesh.nFx)
                e[indActive_Fx] = 1.
                self._Pafx = Utils.sdiag(e)
        return self._Pafx

    @property
    def Pafy(self):
        """
        diagonal matrix that nulls out inactive y-faces
        to full modelling space (ie. nFy x nindActive_Fy )

        :rtype: scipy.sparse.csr_matrix
        :return: active face-y diagonal matrix
        """
        if getattr(self, '_Pafy', None) is None:
            if self.indActive is None:
                self._Pafy = Utils.speye(self.mesh.nFy)
            else:
                indActive_Fy = (self.mesh.aveFy2CC.T * self.indActive) >= 1
                e = np.zeros(self.mesh.nFy)
                e[indActive_Fy] = 1.
                self._Pafy = Utils.sdiag(e)
        return self._Pafy

    @property
    def Pafz(self):
        """
        diagonal matrix that nulls out inactive z-faces
        to full modelling space (ie. nFz x nindActive_Fz )

        :rtype: scipy.sparse.csr_matrix
        :return: active face-z diagonal matrix
        """
        if getattr(self, '_Pafz', None) is None:
            if self.indActive is None:
                self._Pafz = Utils.speye(self.mesh.nFz)
            else:
                indActive_Fz = (self.mesh.aveFz2CC.T * self.indActive) >= 1
                e = np.zeros(self.mesh.nFz)
                e[indActive_Fz] = 1.
                self._Pafz = Utils.sdiag(e)
        return self._Pafz


if __name__ == '__main__':
    from SimPEG import Mesh, np
    mesh = Mesh.TensorMesh([10, 10])
    L = np.ones(mesh.nC)
    src = StreamingCurrents([], L=L, mesh=mesh)
    thing = src.MfLiI
    if thing is not None:
        pass



import numpy as np
import scipy.sparse as sp
import warnings
import properties

from .. import Props
from .. import Utils

###############################################################################
#                                                                             #
#                             Regularization Mesh                             #
#                                                                             #
###############################################################################

class RegularizationMesh(Props.BaseSimPEG):
    """
    **Regularization Mesh**

    This contains the operators used in the regularization. Note that these
    are not necessarily true differential operators, but are constructed from
    a SimPEG Mesh.

    :param discretize.base.BaseMesh mesh: problem mesh
    :param numpy.ndarray indActive: bool array, size nC, that is True where we have active cells. Used to reduce the operators so we regularize only on active cells

    """

    regularization_type = None  # or 'Simple', 'Sparse' or 'Tikhonov'

    def __init__(self, mesh, **kwargs):
        self.mesh = mesh
        Utils.setKwargs(self, **kwargs)

    indActive = properties.Array("active indices in mesh", dtype=[bool, int])

    @properties.validator('indActive')
    def _cast_to_bool(self, change):
        value = change['value']
        if value is not None:
            if value.dtype != 'bool':  # cast it to a bool otherwise
                tmp = value
                value = np.zeros(self.mesh.nC, dtype=bool)
                value[tmp] = True
                change['value'] = value

    @property
    def vol(self):
        """
        reduced volume vector

        :rtype: numpy.ndarray
        :return: reduced cell volume
        """
        if getattr(self, '_vol', None) is None:
            self._vol = self.Pac.T * self.mesh.vol
        return self._vol

    @property
    def nC(self):
        """
        reduced number of cells

        :rtype: int
        :return: number of cells being regularized
        """
        if self.indActive is not None:
            return int(self.indActive.sum())
        return self.mesh.nC

    @property
    def dim(self):
        """
        dimension of regularization mesh (1D, 2D, 3D)

        :rtype: int
        :return: dimension
        """
        if getattr(self, '_dim', None) is None:
            self._dim = self.mesh.dim
        return self._dim

    @property
    def Pac(self):
        """
        projection matrix that takes from the reduced space of active cells to
        full modelling space (ie. nC x nindActive)

        :rtype: scipy.sparse.csr_matrix
        :return: active cell projection matrix
        """
        if getattr(self, '_Pac', None) is None:
            if self.indActive is None:
                self._Pac = Utils.speye(self.mesh.nC)
            else:
                self._Pac = Utils.speye(self.mesh.nC)[:, self.indActive]
        return self._Pac

    @property
    def Pafx(self):
        """
        projection matrix that takes from the reduced space of active x-faces
        to full modelling space (ie. nFx x nindActive_Fx )

        :rtype: scipy.sparse.csr_matrix
        :return: active face-x projection matrix
        """
        if getattr(self, '_Pafx', None) is None:
            if self.indActive is None:
                self._Pafx = Utils.speye(self.mesh.nFx)
            else:
                # if getattr(self.mesh, 'aveCC2Fx', None) is not None:
                if self.mesh._meshType == "TREE":
                    if self.regularization_type == "Tikhonov":
                        indActive_Fx = (
                            (self.mesh.aveFx2CC.T * self.indActive) >= 1
                        )
                        self._Pafx = (
                            Utils.speye(self.mesh.nFx)[:, indActive_Fx]
                        )
                    else:
                        indActive_Fx = (
                            (
                                self.mesh._aveCC2FxStencil() *
                                self.indActive
                            ) >= 1
                        )
                        self._Pafx = (
                            Utils.speye(self.mesh.ntFx)[:, indActive_Fx]
                        )
                else:
                    indActive_Fx = self.mesh.aveFx2CC.T * self.indActive >= 1

                    self._Pafx = Utils.speye(self.mesh.nFx)[:, indActive_Fx]
        return self._Pafx

    @property
    def Pafy(self):
        """
        projection matrix that takes from the reduced space of active y-faces
        to full modelling space (ie. nFy x nindActive_Fy )

        :rtype: scipy.sparse.csr_matrix
        :return: active face-y projection matrix
        """
        if getattr(self, '_Pafy', None) is None:
            if self.indActive is None:
                self._Pafy = Utils.speye(self.mesh.nFy)
            else:
                # if getattr(self.mesh, 'aveCC2Fy', None) is not None:
                if self.mesh._meshType == "TREE":
                    if self.regularization_type == "Tikhonov":
                        indActive_Fy = (
                            (self.mesh.aveFy2CC.T * self.indActive) >= 1
                        )
                        self._Pafy = (
                            Utils.speye(self.mesh.nFy)[:, indActive_Fy]
                        )
                    else:
                        indActive_Fy = (
                            (
                                self.mesh._aveCC2FyStencil() *
                                self.indActive
                            ) >= 1
                        )
                        self._Pafy = (
                            Utils.speye(self.mesh.ntFy)[:, indActive_Fy]
                        )
                else:
                    indActive_Fy = (self.mesh.aveFy2CC.T * self.indActive) >= 1
                    self._Pafy = Utils.speye(self.mesh.nFy)[:, indActive_Fy]
        return self._Pafy

    @property
    def Pafz(self):
        """
        projection matrix that takes from the reduced space of active z-faces
        to full modelling space (ie. nFz x nindActive_Fz )

        :rtype: scipy.sparse.csr_matrix
        :return: active face-z projection matrix
        """
        if getattr(self, '_Pafz', None) is None:
            if self.indActive is None:
                self._Pafz = Utils.speye(self.mesh.nFz)
            else:
                # if getattr(self.mesh, 'aveCC2Fz', None) is not None:
                if self.mesh._meshType == "TREE":
                    if self.regularization_type == "Tikhonov":
                        indActive_Fz = (
                            (self.mesh.aveFz2CC.T * self.indActive) >= 1
                        )
                        self._Pafz = (
                            Utils.speye(self.mesh.nFz)[:, indActive_Fz]
                        )
                    else:
                        indActive_Fz = (
                            (
                                self.mesh._aveCC2FzStencil() *
                                self.indActive
                            ) >= 1
                        )
                        self._Pafz = (
                            Utils.speye(self.mesh.ntFz)[:, indActive_Fz]
                        )
                else:
                    indActive_Fz = (self.mesh.aveFz2CC.T * self.indActive) >= 1
                    self._Pafz = Utils.speye(self.mesh.nFz)[:, indActive_Fz]
        return self._Pafz

    @property
    def aveFx2CC(self):
        """
        averaging from active cell centers to active x-faces

        :rtype: scipy.sparse.csr_matrix
        :return: averaging from active cell centers to active x-faces
        """
        if getattr(self, '_aveFx2CC', None) is None:
            if self.mesh._meshType == "TREE":
                if self.regularization_type == "Tikhonov":
                    self._aveFx2CC = (
                        self.Pac.T * self.mesh.aveFx2CC * self.Pafx
                    )

                else:
                    nCinRow = mkvc((self.aveCC2Fx.T).sum(1))
                    nCinRow[nCinRow > 0] = 1./nCinRow[nCinRow > 0]
                    self._aveFx2CC = (
                        Utils.sdiag(nCinRow) *
                        self.aveCC2Fx.T
                    )

            else:
                self._aveFx2CC = self.Pac.T * self.mesh.aveFx2CC * self.Pafx

        return self._aveFx2CC

    @property
    def aveCC2Fx(self):
        """
        averaging from active x-faces to active cell centers

        :rtype: scipy.sparse.csr_matrix
        :return: averaging matrix from active x-faces to active cell centers
        """
        if getattr(self, '_aveCC2Fx', None) is None:

            # if getattr(self.mesh, 'aveCC2Fx', None) is not None:
            if self.mesh._meshType == "TREE":
                if self.regularization_type == "Tikhonov":
                    self._aveCC2Fx = (
                        Utils.sdiag(1./(self.aveFx2CC.T).sum(1)) *
                        self.aveFx2CC.T
                    )
                else:
                    self._aveCC2Fx = (
                        self.Pafx.T * self.mesh._aveCC2FxStencil() * self.Pac
                    )
            else:
                self._aveCC2Fx = (
                    Utils.sdiag(1./(self.aveFx2CC.T).sum(1)) * self.aveFx2CC.T
                )
        return self._aveCC2Fx

    @property
    def aveFy2CC(self):
        """
        averaging from active cell centers to active y-faces

        :rtype: scipy.sparse.csr_matrix
        :return: averaging from active cell centers to active y-faces
        """
        if getattr(self, '_aveFy2CC', None) is None:
            if self.mesh._meshType == "TREE":
                if self.regularization_type == "Tikhonov":
                    self._aveFy2CC = (
                        self.Pac.T * self.mesh.aveFy2CC * self.Pafy
                    )

                else:
                    nCinRow = mkvc((self.aveCC2Fy.T).sum(1))
                    nCinRow[nCinRow > 0] = 1./nCinRow[nCinRow > 0]
                    self._aveFy2CC = (
                        Utils.sdiag(nCinRow) *
                        self.aveCC2Fy.T
                    )

            else:
                self._aveFy2CC = self.Pac.T * self.mesh.aveFy2CC * self.Pafy

        return self._aveFy2CC

    @property
    def aveCC2Fy(self):
        """
        averaging from active y-faces to active cell centers

        :rtype: scipy.sparse.csr_matrix
        :return: averaging matrix from active y-faces to active cell centers
        """
        if getattr(self, '_aveCC2Fy', None) is None:
            # if getattr(self.mesh, 'aveCC2Fy', None) is not None:
            if self.mesh._meshType == "TREE":
                if self.regularization_type == "Tikhonov":
                    self._aveCC2Fy = (
                        Utils.sdiag(1./(self.aveFy2CC.T).sum(1)) *
                        self.aveFy2CC.T
                    )
                else:
                    self._aveCC2Fy = (
                        self.Pafy.T * self.mesh._aveCC2FyStencil() * self.Pac
                    )
            else:
                self._aveCC2Fy = (
                    Utils.sdiag(1./(self.aveFy2CC.T).sum(1)) * self.aveFy2CC.T
                )
        return self._aveCC2Fy

    @property
    def aveFz2CC(self):
        """
        averaging from active cell centers to active z-faces

        :rtype: scipy.sparse.csr_matrix
        :return: averaging from active cell centers to active z-faces
        """
        if getattr(self, '_aveFz2CC', None) is None:
            if self.mesh._meshType == "TREE":
                if self.regularization_type == "Tikhonov":
                    self._aveFz2CC = (
                        self.Pac.T * self.mesh.aveFz2CC * self.Pafz
                    )

                else:
                    nCinRow = mkvc((self.aveCC2Fz.T).sum(1))
                    nCinRow[nCinRow > 0] = 1./nCinRow[nCinRow > 0]
                    self._aveFz2CC = (
                        Utils.sdiag(nCinRow) *
                        self.aveCC2Fz.T
                    )

            else:
                self._aveFz2CC = self.Pac.T * self.mesh.aveFz2CC * self.Pafz

        return self._aveFz2CC

    @property
    def aveCC2Fz(self):
        """
        averaging from active z-faces to active cell centers

        :rtype: scipy.sparse.csr_matrix
        :return: averaging matrix from active z-faces to active cell centers
        """
        if getattr(self, '_aveCC2Fz', None) is None:
            # if getattr(self.mesh, 'aveCC2Fz', None) is not None:
            if self.mesh._meshType == "TREE":
                if self.regularization_type == "Tikhonov":
                    self._aveCC2Fz = (
                        Utils.sdiag(1./(self.aveFz2CC.T).sum(1)) *
                        self.aveFz2CC.T
                    )
                else:
                    self._aveCC2Fz = (
                        self.Pafz.T * self.mesh._aveCC2FzStencil() * self.Pac
                    )
            else:
                self._aveCC2Fz = (
                    Utils.sdiag(1./(self.aveFz2CC.T).sum(1)) * self.aveFz2CC.T
                )
        return self._aveCC2Fz

    @property
    def cellDiffx(self):
        """
        cell centered difference in the x-direction

        :rtype: scipy.sparse.csr_matrix
        :return: differencing matrix for active cells in the x-direction
        """
        if getattr(self, '_cellDiffx', None) is None:
            self._cellDiffx = self.Pafx.T * self.mesh.cellGradx * self.Pac
        return self._cellDiffx

    @property
    def cellDiffy(self):
        """
        cell centered difference in the y-direction

        :rtype: scipy.sparse.csr_matrix
        :return: differencing matrix for active cells in the y-direction
        """
        if getattr(self, '_cellDiffy', None) is None:
            self._cellDiffy = self.Pafy.T * self.mesh.cellGrady * self.Pac
        return self._cellDiffy

    @property
    def cellDiffz(self):
        """
        cell centered difference in the z-direction

        :rtype: scipy.sparse.csr_matrix
        :return: differencing matrix for active cells in the z-direction
        """
        if getattr(self, '_cellDiffz', None) is None:
            self._cellDiffz = self.Pafz.T * self.mesh.cellGradz * self.Pac
        return self._cellDiffz

    @property
    def faceDiffx(self):
        """
        x-face differences

        :rtype: scipy.sparse.csr_matrix
        :return: differencing matrix for active faces in the x-direction
        """
        if getattr(self, '_faceDiffx', None) is None:
            self._faceDiffx = self.Pac.T * self.mesh.faceDivx * self.Pafx
        return self._faceDiffx

    @property
    def faceDiffy(self):
        """
        y-face differences

        :rtype: scipy.sparse.csr_matrix
        :return: differencing matrix for active faces in the y-direction
        """
        if getattr(self, '_faceDiffy', None) is None:
            self._faceDiffy = self.Pac.T * self.mesh.faceDivy * self.Pafy
        return self._faceDiffy

    @property
    def faceDiffz(self):
        """
        z-face differences

        :rtype: scipy.sparse.csr_matrix
        :return: differencing matrix for active faces in the z-direction
        """
        if getattr(self, '_faceDiffz', None) is None:
            self._faceDiffz = self.Pac.T * self.mesh.faceDivz * self.Pafz
        return self._faceDiffz

    @property
    def cellDiffxStencil(self):
        """
        cell centered difference stencil (no cell lengths include) in the
        x-direction

        :rtype: scipy.sparse.csr_matrix
        :return: differencing matrix for active cells in the x-direction
        """
        if getattr(self, '_cellDiffxStencil', None) is None:

            self._cellDiffxStencil = (
                self.Pafx.T * self.mesh._cellGradxStencil * self.Pac
            )
        return self._cellDiffxStencil

    @property
    def cellDiffyStencil(self):
        """
        cell centered difference stencil (no cell lengths include) in the
        y-direction

        :rtype: scipy.sparse.csr_matrix
        :return: differencing matrix for active cells in the y-direction
        """
        if self.dim < 2:
            return None
        if getattr(self, '_cellDiffyStencil', None) is None:

            self._cellDiffyStencil = (
                self.Pafy.T * self.mesh._cellGradyStencil * self.Pac
            )
        return self._cellDiffyStencil

    @property
    def cellDiffzStencil(self):
        """
        cell centered difference stencil (no cell lengths include) in the
        y-direction

        :rtype: scipy.sparse.csr_matrix
        :return: differencing matrix for active cells in the y-direction
        """
        if self.dim < 3:
            return None
        if getattr(self, '_cellDiffzStencil', None) is None:

            self._cellDiffzStencil = (
                self.Pafz.T * self.mesh._cellGradzStencil * self.Pac
            )
        return self._cellDiffzStencil




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="geofem", # Replace with your own username
    version="v0.0.5",
    author="Isadora",
    author_email="isadora.s.macedo@gmail.com",
    description="Modelagem geofísica direta de campos eletromagnéticos via volumes finitos: MT e MCSEM.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/iisadoramacedo/geofem-master",
    download_url="https://github.com/iisadoramacedo/geofem-master/archive/v0.0.5.tar.gz",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
    ],
    install_requires=[
        "numpy>=1.7",
        "scipy>=1.4.0",
        "pymatsolver>=0.1.1",
        "matplotlib",
        "properties>=0.5.2",
        "vectormath>=0.2.0",
        "discretize>=0.4.0",
        "geoana>=0.0.4",
        "empymod>=2.0.0",
        "pandas",
        "numba>=0.45.0",
    ],

    python_requires='>=3.7',
)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import SimPEG
from SimPEG.Utils import Identity, Zero
import numpy as np
from scipy.constants import epsilon_0


class FieldsDC(SimPEG.Problem.Fields):
    knownFields = {}
    dtype = float

    def _phiDeriv(self, src, du_dm_v, v, adjoint=False):
        if (
            getattr(self, '_phiDeriv_u', None) is None or
            getattr(self, '_phiDeriv_m', None) is None
        ):
            raise NotImplementedError(
                'Getting phiDerivs from {0!s} is not '
                'implemented'.format(self.knownFields.keys()[0])
            )

        if adjoint:
            return (self._phiDeriv_u(src, v, adjoint=adjoint),
                    self._phiDeriv_m(src, v, adjoint=adjoint))

        return (np.array(self._phiDeriv_u(src, du_dm_v, adjoint) +
                         self._phiDeriv_m(src, v, adjoint), dtype=float))

    def _eDeriv(self, src, du_dm_v, v, adjoint=False):
        if (
            getattr(self, '_eDeriv_u', None) is None or
            getattr(self, '_eDeriv_m', None) is None
        ):
            raise NotImplementedError(
                'Getting eDerivs from {0!s} is not '
                'implemented'.format(self.knownFields.keys()[0])
            )

        if adjoint:
            return (self._eDeriv_u(src, v, adjoint),
                    self._eDeriv_m(src, v, adjoint))
        return (np.array(self._eDeriv_u(src, du_dm_v, adjoint) +
                         self._eDeriv_m(src, v, adjoint), dtype=float))

    def _jDeriv(self, src, du_dm_v, v, adjoint=False):
        if (
            getattr(self, '_jDeriv_u', None) is None or
            getattr(self, '_jDeriv_m', None) is None
        ):
            raise NotImplementedError(
                'Getting jDerivs from {0!s} is not '
                'implemented'.format(self.knownFields.keys()[0])
            )

        if adjoint:
            return (self._jDeriv_u(src, v, adjoint),
                    self._jDeriv_m(src, v, adjoint))
        return (np.array(self._jDeriv_u(src, du_dm_v, adjoint) +
                         self._jDeriv_m(src, v, adjoint), dtype=float))


class Fields_CC(FieldsDC):
    knownFields = {'phiSolution': 'CC'}
    aliasFields = {
        'phi': ['phiSolution', 'CC', '_phi'],
        'j': ['phiSolution', 'F', '_j'],
        'e': ['phiSolution', 'F', '_e'],
        'charge': ['phiSolution', 'CC', '_charge'],
        'charge_density': ['phiSolution', 'CC', '_charge_density'],
    }
    # primary - secondary
    # CC variables

    def __init__(self, mesh, survey, **kwargs):
        FieldsDC.__init__(self, mesh, survey, **kwargs)

        if getattr(self.survey.prob, 'bc_type', None) == 'Dirichlet':
            self.cellGrad = -mesh.faceDiv.T
        elif getattr(self.survey.prob, 'bc_type', None) == 'Neumann':
            if self.mesh._meshType == "TREE":
                raise NotImplementedError()
            mesh.setCellGradBC("neumann")
            self.cellGrad = mesh.cellGrad
        else:
            mesh.setCellGradBC("neumann")
            self.cellGrad = mesh.cellGrad

    def startup(self):
        # self.prob = self.survey.prob
        self._MfRhoI = self.survey.prob.MfRhoI
        self._MfRhoIDeriv = self.survey.prob.MfRhoIDeriv
        self._MfRho = self.survey.prob.MfRho
        self._aveF2CCV = self.survey.prob.mesh.aveF2CCV
        self._nC = self.survey.prob.mesh.nC
        self._Grad = self.survey.prob.Grad
        self._MfI = self.survey.prob.MfI
        self._Vol = self.survey.prob.Vol
        self._faceDiv = self.survey.prob.mesh.faceDiv

    def _GLoc(self, fieldType):
        if fieldType == 'phi':
            return 'CC'
        elif fieldType == 'e' or fieldType == 'j':
            return 'F'
        else:
            raise Exception('Field type must be phi, e, j')

    def _phi(self, phiSolution, srcList):
        return phiSolution

    def _phiDeriv_u(self, src, v, adjoint=False):
        return Identity()*v

    def _phiDeriv_m(self, src, v, adjoint=False):
        return Zero()

    def _j(self, phiSolution, srcList):
        """
            .. math::

                \mathbf{j} = \mathbf{M}^{f \ -1}_{\rho} \mathbf{G} \phi
        """
        return self._MfRhoI*self._Grad*phiSolution

    def _jDeriv_u(self, src, v, adjoint=False):
        if adjoint:
            return self._Grad.T * (self._MfRhoI.T * v)
        return self._MfRhoI * (self._Grad * v)

    def _jDeriv_m(self, src, v, adjoint=False):
        if adjoint:
            return self._Grad.T * self._MfRhoIDeriv(v, adjoint=True)
        return self._MfRhoIDeriv(self._Grad * v)

    def _e(self, phiSolution, srcList):
        """
            .. math::

                \vec{e} = \rho \vec{j}
        """
        # return self._MfI * self._MfRho * self._j(phiSolution, srcList)
        return self._MfI * self._Grad * phiSolution
        # prob._MfI * cart_mesh.faceDiv.T * p

    def _eDeriv_u(self, src, v, adjoint=False):
        if adjoint:
            return self._Grad.T * (self._MfI.T * v)
        return self._MfI * (self._Grad * v)

    def _eDeriv_m(self, src, v, adjoint=False):
        return Zero()

    def _charge(self, phiSolution, srcList):
        """
            .. math::

                \int \nabla \codt \vec{e} =  \int \frac{\rho_v }{\epsillon_0}
        """
        return epsilon_0*self._Vol*(
            self._faceDiv*self._e(phiSolution, srcList)
        )

    def _charge_density(self, phiSolution, srcList):
        """
            .. math::

                \frac{1}{V}\int \nabla \codt \vec{e} =
                \frac{1}{V}\int \frac{\rho_v }{\epsillon_0}
        """
        return epsilon_0*(
            self._faceDiv*self._e(phiSolution, srcList)
        )


class Fields_N(FieldsDC):
    knownFields = {'phiSolution': 'N'}
    aliasFields = {
        'phi': ['phiSolution', 'N', '_phi'],
        'j': ['phiSolution', 'E', '_j'],
        'e': ['phiSolution', 'E', '_e'],
        'charge_density': ['phiSolution', 'N', '_charge_density'],
        'charge': ['phiSolution', 'N', '_charge'],
    }
    # primary - secondary
    # N variables

    def __init__(self, mesh, survey, **kwargs):
        FieldsDC.__init__(self, mesh, survey, **kwargs)

    def startup(self):
        self.prob = self.survey.prob

    def _GLoc(self, fieldType):
        if fieldType == 'phi':
            return 'N'
        elif fieldType == 'e' or fieldType == 'j':
            return 'E'
        else:
            raise Exception('Field type must be phi, e, j')

    def _phi(self, phiSolution, srcList):
        return phiSolution

    def _phiDeriv_u(self, src, v, adjoint=False):
        return Identity()*v

    def _phiDeriv_m(self, src, v, adjoint=False):
        return Zero()

    def _j(self, phiSolution, srcList):
        """
            In EB formulation j is not well-defined!!
            .. math::
                \mathbf{j} = - \mathbf{M}^{e}_{\sigma} \mathbf{G} \phi
        """
        return self.prob.MeI * self.prob.MeSigma * self._e(phiSolution, srcList)

    def _e(self, phiSolution, srcList):
        """
            In HJ formulation e is not well-defined!!
            .. math::
                \vec{e} = -\nabla \phi
        """
        return -self.mesh.nodalGrad * phiSolution

    def _charge(self, phiSolution, srcList):
        """
            .. math::
                \int \nabla \codt \vec{e} =  \int \frac{\rho_v }{\epsillon_0}
        """
        return -epsilon_0*(
            self.mesh.nodalGrad.T *
            self.mesh.getEdgeInnerProduct() *
            self._e(phiSolution, srcList)
        )

    def _charge_density(self, phiSolution, srcList):
        return self._charge(phiSolution, srcList) / self.mesh.vol



import numpy as np
from scipy.spatial import Delaunay
from SimPEG import Utils, Mesh


def PolygonInd(mesh, pts):
    # if mesh.dim == 2:
    hull = Delaunay(pts)
    inds = hull.find_simplex(mesh.gridCC) >= 0
    # else:
    return inds


def readSeepageModel(fname, mesh=None, xsurf=None, ysurf=None):

    fluiddata = pd.read_csv(fname)
    header = fluiddata.keys()
    xyz = np.c_[fluiddata['X (m)'].values, fluiddata['Y (m)'].values]
    h = fluiddata['Total Head (m)'].values
    Ux = fluiddata['X-Velocity Magnitude (m/sec)'].values
    Uy = fluiddata['Y-Velocity Magnitude (m/sec)'].values
    Gradx = fluiddata['X-Gradient'].values
    Grady = fluiddata['Y-Gradient'].values

    Pressure = fluiddata['Pressure Head (m)'].values
    Sw = fluiddata[fluiddata.keys()[17]].values
    Kx = fluiddata["X-Conductivity (m/sec)"]
    Ky = fluiddata["Y-Conductivity (m/sec)"]

    if mesh is None:
        # TODO: this is a specific set up ...
        xmin, ymin = xyz[:, 0].min(), xyz[:, 1].min()
        cs = 0.5
        ncx = 152*2
        ncy = 36*2
        npad = 5
        hx = [(cs, npad, -1.3), (cs, ncx), (cs, npad, 1.3)]
        hy = [(cs, npad, -1.3), (cs, ncy)]
        mesh = Mesh.TensorMesh([hx, hy], x0=[xmin, ymin])
        mesh._x0 = np.r_[xmin-mesh.hx[:10].sum(), xmin-mesh.hy[:10].sum()]
        # ...
        xsurf = np.r_[-1e10, 55, 90, 94, 109, 112, 126.5, +1e10]
        ysurf = np.r_[27.5, 27.5, 43.2, 43.2, 35, 35, 27.5, 27.5]
        yup = np.ones_like(ysurf)*45
        actind = Utils.surface2ind_topo(mesh, np.c_[xsurf, ysurf])
        waterheight = 40.
        waterind = (
            (np.logical_and(~actind, mesh.gridCC[:, 1] < 40.)) &
            (mesh.gridCC[:, 0] < 90.)
        )

    F_hlin = LinearNDInterpolator(xyz, h)
    hccIn = F_hlin(mesh.gridCC[actind, :])
    F_hnear = NearestNDInterpolator(mesh.gridCC[actind, :], hccIn)
    hcc = F_hnear(mesh.gridCC)
    fluiddata = {
        "xyz": xyz, "h": h, "Sw": Sw,
        "Kx": Kx, "Ky": Ky, "P": Pressure, "Ux": Ux, "Uy": Uy,
        "Gradx": Gradx, "Grady": Grady,
        "hcc":  hcc, "mesh": mesh,
        "actind": actind,
        "waterind":  waterind,
        "xsurf": xsurf, "ysurf": ysurf, "yup": yup
    }

    return fluiddata


def writeVectorUBC(mesh, fileName, model):
    """
        Writes a vector model associated with a SimPEG TensorMesh
        to a UBC-GIF format model file.

        :param string fileName: File to write to
        :param numpy.ndarray model: The model
    """

    modelMatTR = np.zeros_like(model)

    for ii in range(3):
        # Reshape model to a matrix
        modelMat = mesh.r(model[:, ii], 'CC', 'CC', 'M')
        # Transpose the axes
        modelMatT = modelMat.transpose((2, 0, 1))
        # Flip UBC order
        modelMatTR[:, ii] = Utils.mkvc(modelMatT[::-1, :, :])

        # Flip z to positive down for MeshTools3D
        if ii == 2:
            modelMatTR[:, ii] *= -1
    np.savetxt(fileName, modelMatTR)


def readVectorUBC(mesh, fileName):
    """Read UBC 3DVector model and generate 3D Vector mesh model

    Input:
    :param string fileName: path to the UBC GIF mesh file to read

    Output:
    :rtype: numpy.ndarray
    :return: model with TensorMesh ordered x3 nC
    """
    model = np.loadtxt(fileName)
    # Fist line is the size of the model
    # model = np.array(model.ravel()[0].split(), dtype=float)

    vx = np.reshape(model[:, 0], (mesh.nCz, mesh.nCx, mesh.nCy), order='F')
    vx = vx[::-1, :, :]
    vx = np.transpose(vx, (1, 2, 0))
    vx = Utils.mkvc(vx)

    vy = np.reshape(model[:, 1], (mesh.nCz, mesh.nCx, mesh.nCy), order='F')
    vy = vy[::-1, :, :]
    vy = np.transpose(vy, (1, 2, 0))
    vy = Utils.mkvc(vy)

    vz = np.reshape(model[:, 2], (mesh.nCz, mesh.nCx, mesh.nCy), order='F')
    vz = vz[::-1, :, :]
    vz = np.transpose(vz, (1, 2, 0))
    vz = Utils.mkvc(vz)

    # Flip z to positive up from MeshTools3D to SimPEG
    model = np.r_[vx, vy, -vz]
    return model



import numpy as np
from scipy.special import ellipk, ellipe
from scipy.constants import mu_0

import properties

from discretize.base import BaseMesh
from discretize import TensorMesh

from SimPEG import Utils


orientationDict = {'X': np.r_[1., 0., 0.],
                   'Y': np.r_[0., 1., 0.],
                   'Z': np.r_[0., 0., 1.]}


def MagneticDipoleVectorPotential(srcLoc, obsLoc, component, moment=1.,
                                  orientation=np.r_[0., 0., 1.],
                                  mu=mu_0):
    """
        Calculate the vector potential of a set of magnetic dipoles
        at given locations 'ref. <http://en.wikipedia.org/wiki/Dipole#Magnetic_vector_potential>'

        :param numpy.ndarray srcLoc: Location of the source(s) (x, y, z)
        :param numpy.ndarray,discretize obsLoc: Where the potentials will be
                                                 calculated (x, y, z) or a
                                                 SimPEG Mesh
        :param str,list component: The component to calculate - 'x', 'y', or
                                   'z' if an array, or grid type if mesh, can
                                   be a list
        :param numpy.ndarray orientation: The vector dipole moment
        :rtype: numpy.ndarray
        :return: The vector potential each dipole at each observation location
    """
    # TODO: break this out!

    if isinstance(orientation, str):
        orientation = orientationDict[orientation]

    assert np.linalg.norm(orientation, 2) == 1., ("orientation must "
                                                            "be a unit vector")

    if type(component) in [list, tuple]:
        out = list(range(len(component)))
        for i, comp in enumerate(component):
            out[i] = MagneticDipoleVectorPotential(srcLoc, obsLoc, comp,
                                                   orientation=orientation,
                                                   mu=mu)
        return np.concatenate(out)

    if isinstance(obsLoc, BaseMesh):
        mesh = obsLoc
        assert component in ['Ex', 'Ey', 'Ez', 'Fx', 'Fy', 'Fz'], ("Components"
                                 "must be in: ['Ex','Ey','Ez','Fx','Fy','Fz']")
        return MagneticDipoleVectorPotential(srcLoc, getattr(mesh, 'grid' +
                                                             component),
                                             component[1],
                                             orientation=orientation)

    if component == 'x':
        dimInd = 0
    elif component == 'y':
        dimInd = 1
    elif component == 'z':
        dimInd = 2
    else:
        raise ValueError('Invalid component')

    srcLoc = np.atleast_2d(srcLoc)
    obsLoc = np.atleast_2d(obsLoc)
    orientation = np.atleast_2d(orientation)

    nObs = obsLoc.shape[0]
    nSrc = srcLoc.shape[0]

    m = moment*np.array(orientation).repeat(nObs, axis=0)
    A = np.empty((nObs, nSrc))
    for i in range(nSrc):
        dR = obsLoc - srcLoc[i, np.newaxis].repeat(nObs, axis=0)
        mCr = np.cross(m, dR)
        r = np.sqrt((dR**2).sum(axis=1))
        A[:, i] = +(mu/(4*np.pi)) * mCr[:, dimInd]/(r**3)
    if nSrc == 1:
        return A.flatten()
    return A


def MagneticDipoleFields(
    srcLoc, obsLoc, component, orientation='Z', moment=1., mu=mu_0
):
    """
        Calculate the vector potential of a set of magnetic dipoles
        at given locations 'ref. <http://en.wikipedia.org/wiki/Dipole#Magnetic_vector_potential>'

        .. math::

            B = \frac{\mu_0}{4 \pi r^3} \left( \frac{3 \vec{r} (\vec{m} \cdot
                                                                \vec{r})}{r^2})
                                                - \vec{m}
                                        \right) \cdot{\hat{rx}}

        :param numpy.ndarray srcLoc: Location of the source(s) (x, y, z)
        :param numpy.ndarray obsLoc: Where the potentials will be calculated
                                     (x, y, z)
        :param str component: The component to calculate - 'x', 'y', or 'z'
        :param numpy.ndarray moment: The vector dipole moment (vertical)
        :rtype: numpy.ndarray
        :return: The vector potential each dipole at each observation location
    """

    if isinstance(orientation, str):
        assert orientation.upper() in ['X', 'Y', 'Z'], ("orientation must be 'x', "
                                                      "'y', or 'z' or a vector"
                                                      "not {}".format(orientation)
                                                      )
    elif (not np.allclose(np.r_[1., 0., 0.], orientation) or
          not np.allclose(np.r_[0., 1., 0.], orientation) or
          not np.allclose(np.r_[0., 0., 1.], orientation)):
        warnings.warn('Arbitrary trasnmitter orientations ({}) not thouroughly tested '
                      'Pull request on a test anyone? bueller?').format(orientation)

    if isinstance(component, str):
        assert component.upper() in ['X', 'Y', 'Z'], ("component must be 'x', "
                                                      "'y', or 'z' or a vector"
                                                      "not {}".format(component)
                                                      )
    elif (not np.allclose(np.r_[1., 0., 0.], component) or
          not np.allclose(np.r_[0., 1., 0.], component) or
          not np.allclose(np.r_[0., 0., 1.], component)):
        warnings.warn('Arbitrary receiver orientations ({}) not thouroughly tested '
                      'Pull request on a test anyone? bueller?').format(component)

    if isinstance(orientation, str):
        orientation = orientationDict[orientation.upper()]

    if isinstance(component, str):
        component = orientationDict[component.upper()]

    assert np.linalg.norm(orientation, 2) == 1., ('orientation must be a unit '
                                                  'vector. Use "moment=X to '
                                                  'scale source fields')

    if np.linalg.norm(component, 2) != 1.:
        warnings.warn('The magnitude of the receiver component vector is > 1, '
                      ' it is {}. The receiver fields will be scaled.'
                      ).format(np.linalg.norm(component, 2))

    srcLoc = np.atleast_2d(srcLoc)
    component = np.atleast_2d(component)
    obsLoc = np.atleast_2d(obsLoc)
    orientation = np.atleast_2d(orientation)

    nObs = obsLoc.shape[0]
    nSrc = int(srcLoc.size / 3.)

    # use outer product to construct an array of [x_src, y_src, z_src]

    m = moment*orientation.repeat(nObs, axis=0)
    B = []

    for i in range(nSrc):
        srcLoc = srcLoc[i, np.newaxis].repeat(nObs, axis=0)
        rx = component.repeat(nObs, axis=0)
        dR = obsLoc - srcLoc
        r = np.sqrt((dR**2).sum(axis=1))
        # mult each element and sum along the axis (vector dot product)
        m_dot_dR_div_r2 = (m * dR).sum(axis=1) / (r**2)

        #multiply the scalar m_dot_dR by the 3D vector r

        rvec_m_dot_dR_div_r2 = np.vstack(
            [np.multiply(m_dot_dR_div_r2, dR[:, i]) for i in range(3)]
        ).T

        # print((3. * rvec_m_dot_dR_div_r2).shape,rvec_m_dot_dR_div_r2.shape, m.shape)
        inside = (3. * rvec_m_dot_dR_div_r2) - m

        # dot product with rx orientation
        inside_dot_rx = (inside * rx).sum(axis=1)
        front = (mu/(4.* np.pi * r**3))
        B.append(Utils.mkvc(np.multiply(front, inside_dot_rx)))

    return np.vstack(B).T


    #     if np.all(orientation == np.r_[1., 0., 0.]):

    #     elif np.all(orientation == np.r_[0., 0., 1.]):
    #         x1 = dR[:, 2]
    #         x2 = dR[:, 0]
    #         x3 = dR[:, 1]


    #     if component == 'x':
    #         B[:, i] = front * (3*x1*x2/r**2)
    #     elif component == 'y':
    #         B[:, i] = front * (3*x1*x3/r**2)
    #     elif component == 'z':
    #         B[:, i] = front * (3*x1**2/r**2-1)
    #     else:
    #         raise Exception("Not Implemented")
    # if nSrc == 1:
    #     return B.flatten()
    # return B



def MagneticLoopVectorPotential(srcLoc, obsLoc, component, radius, orientation='Z', mu=mu_0):
    """
        Calculate the vector potential of horizontal circular loop
        at given locations

        :param numpy.ndarray srcLoc: Location of the source(s) (x, y, z)
        :param numpy.ndarray,discretize obsLoc: Where the potentials will be calculated (x, y, z) or a SimPEG Mesh
        :param str,list component: The component to calculate - 'x', 'y', or 'z' if an array, or grid type if mesh, can be a list
        :param numpy.ndarray I: Input current of the loop
        :param numpy.ndarray radius: radius of the loop
        :rtype: numpy.ndarray
        :return: The vector potential each dipole at each observation location
    """

    if isinstance(orientation, str):
        if orientation.upper() != 'Z':
            raise NotImplementedError('Only Z oriented loops implemented')
    elif not np.allclose(orientation, np.r_[0., 0., 1.]):
        raise NotImplementedError('Only Z oriented loops implemented')

    if type(component) in [list, tuple]:
        out = list(range(len(component)))
        for i, comp in enumerate(component):
            out[i] = MagneticLoopVectorPotential(srcLoc, obsLoc, comp, radius,
                                                 orientation, mu)
        return np.concatenate(out)

    if isinstance(obsLoc, BaseMesh):
        mesh = obsLoc
        assert component in ['Ex','Ey','Ez','Fx','Fy','Fz'], "Components must be in: ['Ex','Ey','Ez','Fx','Fy','Fz']"
        return MagneticLoopVectorPotential(srcLoc, getattr(mesh,'grid'+component), component[1], radius, mu)

    srcLoc = np.atleast_2d(srcLoc)
    obsLoc = np.atleast_2d(obsLoc)

    n = obsLoc.shape[0]
    nSrc = srcLoc.shape[0]

    if component=='z':
        A = np.zeros((n, nSrc))
        if nSrc ==1:
            return A.flatten()
        return A

    else:

        A = np.zeros((n, nSrc))
        for i in range (nSrc):
            x = obsLoc[:, 0] - srcLoc[i, 0]
            y = obsLoc[:, 1] - srcLoc[i, 1]
            z = obsLoc[:, 2] - srcLoc[i, 2]
            r = np.sqrt(x**2 + y**2)
            m = (4 * radius * r) / ((radius + r)**2 + z**2)
            m[m > 1.] = 1.
            # m might be slightly larger than 1 due to rounding errors
            # but ellipke requires 0 <= m <= 1
            K = ellipk(m)
            E = ellipe(m)
            ind = (r > 0) & (m < 1)
            # % 1/r singular at r = 0 and K(m) singular at m = 1
            Aphi = np.zeros(n)
            # % Common factor is (mu * I) / pi with I = 1 and mu = 4e-7 * pi.
            Aphi[ind] = ((mu / (np.pi * np.sqrt(m[ind])) *
                         np.sqrt(radius / r[ind]) *((1. - m[ind] / 2.) *
                         K[ind] - E[ind])))
            if component == 'x':
                A[ind, i] = Aphi[ind] * (-y[ind] / r[ind] )
            elif component == 'y':
                A[ind, i] = Aphi[ind] * ( x[ind] / r[ind] )
            else:
                raise ValueError('Invalid component')

        if nSrc == 1:
            return A.flatten()
        return A

if __name__ == '__main__':
    from SimPEG import Mesh
    import matplotlib.pyplot as plt
    cs = 20
    ncx, ncy, ncz = 41, 41, 40
    hx = np.ones(ncx)*cs
    hy = np.ones(ncy)*cs
    hz = np.ones(ncz)*cs
    mesh = TensorMesh([hx, hy, hz], 'CCC')
    srcLoc = np.r_[0., 0., 0.]
    Ax = MagneticLoopVectorPotential(srcLoc, mesh.gridEx, 'x', 200)
    Ay = MagneticLoopVectorPotential(srcLoc, mesh.gridEy, 'y', 200)
    Az = MagneticLoopVectorPotential(srcLoc, mesh.gridEz, 'z', 200)
    A = np.r_[Ax, Ay, Az]
    B0 = mesh.edgeCurl*A
    J0 = mesh.edgeCurl.T*B0

    # mesh.plotImage(A, vType = 'Ex')
    # mesh.plotImage(A, vType = 'Ey')

    mesh.plotImage(B0, vType = 'Fx')
    mesh.plotImage(B0, vType = 'Fy')
    mesh.plotImage(B0, vType = 'Fz')

    # # mesh.plotImage(J0, vType = 'Ex')
    # mesh.plotImage(J0, vType = 'Ey')
    # mesh.plotImage(J0, vType = 'Ez')

    plt.show()





from __future__ import print_function

# Mandatory modules
import sys
import time
import numpy
import scipy
import textwrap
import platform
import multiprocessing

# Required packages
import SimPEG
import cython
import properties
import vectormath
import discretize
import pymatsolver

# Optional modules
try:
    import IPython
    from IPython.display import HTML, Pretty
except ImportError:
    IPython = False
try:
    import matplotlib
except ImportError:
    matplotlib = False
try:
    import ipywidgets
except ImportError:
    ipywidgets = False
try:
    import numexpr
except ImportError:
    numexpr = False
try:
    import mkl
except ImportError:
    mkl = False

# Get mkl info from numexpr or mkl, if available
if mkl:
    mklinfo = mkl.get_version_string()
elif numexpr:
    mklinfo = numexpr.get_vml_version()
else:
    mklinfo = False


class Versions:
    """Print date, time, and version information.

    Print date, time, and package version information in any environment
    (Jupyter notebook, IPython console, Python console, QT console), either as
    html-table (notebook) or as plain text (anywhere).

    Always shown are the OS, number of CPU(s), ``numpy``, ``scipy``,
    ``SimPEG``, ``cython``, ``properties``, ``vectormath``, ``discretize``,
    ``pymatsolver``, ``sys.version``, and time/date.

    Additionally shown are, if they can be imported, ``IPython``,
    ``matplotlib``, and ``ipywidgets``. It also shows MKL information, if
    available.

    All modules provided in ``add_pckg`` are also shown. They have to be
    imported before ``Versions`` is called.

    This script is an adapted version of ``empymod.printversions``,
    (https://empymod.github.io) which itself was heavily inspired by

        - ipynbtools.py from qutip https://github.com/qutip
        - watermark.py from https://github.com/rasbt/watermark


    **Parameters**

    add_pckg : packages, optional
        Package or list of packages to add to output information (must be
        imported beforehand).

    ncol : int, optional
        Number of package-columns in html table; only has effect if
        ``mode='HTML'`` or ``mode='html'``. Defaults to 3.


    **Examples**

    >>> import pytest
    >>> import dateutil
    >>> from SimPEG import Versions
    >>> Versions()                 # Default values
    >>> Versions(pytest)           # Provide additional package
    >>> Versions([pytest, dateutil], ncol=5)  # Define nr of columns

    """

    def __init__(self, add_pckg=None, ncol=4):
        """Initiate and add packages and number of columns to self."""
        self.add_pckg = self._get_packages(add_pckg)
        self.ncol = int(ncol)

    def __repr__(self):
        r"""Plain-text version information."""

        # Width for text-version
        n = 54
        text = u'\n' + ''.join(n*['-']) + '\n'

        # Date and time info as title
        text += time.strftime('  %a %b %d %H:%M:%S %Y %Z\n\n')

        # OS and CPUs
        text += '{:>15}'.format(platform.system())+' : OS\n'
        text += '{:>15}'.format(multiprocessing.cpu_count())+' : CPU(s)\n'

        # Loop over packages
        for pckg in self.add_pckg:
            text += '{:>15} : {}\n'.format(pckg.__version__, pckg.__name__)

        # sys.version
        text += '\n'
        for txt in textwrap.wrap(sys.version, n-4):
            text += '  '+txt+'\n'

        # mkl version
        if mklinfo:
            text += '\n'
            for txt in textwrap.wrap(mklinfo, n-4):
                text += '  '+txt+'\n'

        # Finish
        text += ''.join(n*['-'])
        return text

    def _repr_html_(self):
        """HTML-rendered version information."""

        # Define html-styles
        border = "border: 2px solid #fff;'"

        def colspan(html, txt, ncol, nrow):
            """Print txt in a row spanning whole table."""
            html += "  <tr>\n"
            html += "     <td style='text-align: center; "
            if nrow == 0:
                html += "font-weight: bold; font-size: 1.2em; "
            elif nrow % 2 == 0:
                html += "background-color: #ddd;"
            html += border + " colspan='"
            html += str(2*ncol)+"'>%s</td>\n" % txt
            html += "  </tr>\n"
            return html

        def cols(html, version, name, ncol, i):
            """Print package information in two cells."""

            # Check if we have to start a new row
            if i > 0 and i % ncol == 0:
                html += "  </tr>\n"
                html += "  <tr>\n"

            html += "    <td style='text-align: right; background-color: #ccc;"
            html += " " + border + ">%s</td>\n" % version

            html += "    <td style='text-align: left; "
            html += border + ">%s</td>\n" % name

            return html, i+1

        # Start html-table
        html = "<table style='border: 3px solid #ddd;'>\n"

        # Date and time info as title
        html = colspan(html, time.strftime('%a %b %d %H:%M:%S %Y %Z'),
                       self.ncol, 0)

        # OS and CPUs
        html += "  <tr>\n"
        html, i = cols(html, platform.system(), 'OS', self.ncol, 0)
        html, i = cols(html, multiprocessing.cpu_count(), 'CPU(s)',
                       self.ncol, i)

        # Loop over packages
        for pckg in self.add_pckg:
            html, i = cols(html, pckg.__version__, pckg.__name__, self.ncol, i)
        # Fill up the row
        while i % self.ncol != 0:
            html += "    <td style= " + border + "></td>\n"
            html += "    <td style= " + border + "></td>\n"
            i += 1
        # Finish row
        html += "  </tr>\n"

        # sys.version
        html = colspan(html, sys.version, self.ncol, 1)

        # mkl version
        if mklinfo:
            html = colspan(html, mklinfo, self.ncol, 2)

        # Finish table
        html += "</table>"

        return html

    @staticmethod
    def _get_packages(add_pckg):
        """Create list of packages."""

        # Mandatory packages
        pckgs = [numpy, scipy, SimPEG, cython, properties, vectormath,
                 discretize, pymatsolver]

        # Optional packages
        for module in [IPython, ipywidgets, matplotlib]:
            if module:
                pckgs += [module]

        # Cast and add add_pckg
        if add_pckg is not None:

            # Cast add_pckg
            if isinstance(add_pckg, tuple):
                add_pckg = list(add_pckg)

            if not isinstance(add_pckg, list):
                add_pckg = [add_pckg, ]

            # Add add_pckg
            pckgs += add_pckg

        return pckgs



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import SimPEG
import numpy as np
from SimPEG.Utils import closestPoints, sdiag
import properties

# Trapezoidal integration for 2D DC problem
def IntTrapezoidal(kys, Pf, y=0.):
    phi = np.zeros(Pf.shape[0])
    nky = kys.size
    dky = np.diff(kys)
    dky = np.r_[dky[0], dky]
    phi0 = 1./np.pi*Pf[:, 0]
    for iky in range(nky):
        phi1 = 1./np.pi*Pf[:, iky]
        phi += phi1*dky[iky]/2.*np.cos(kys[iky]*y)
        phi += phi0*dky[iky]/2.*np.cos(kys[iky]*y)
        phi0 = phi1.copy()
    return phi

# Receiver classes
class BaseRx(SimPEG.Survey.BaseRx):
    """
    Base DC receiver
    """
    locs = None
    rxType = None

    data_type = properties.StringChoice(
        "Type of DC-IP survey",
        required=True,
        default="volt",
        choices=[
           "volt",
           "apparent_resistivity",
           "apparent_chargeability"
        ]
    )

    data_type = 'volt'

    knownRxTypes = {
        'phi': ['phi', None],
        'ex': ['e', 'x'],
        'ey': ['e', 'y'],
        'ez': ['e', 'z'],
        'jx': ['j', 'x'],
        'jy': ['j', 'y'],
        'jz': ['j', 'z'],
    }

    def __init__(self, locs, rxType, **kwargs):
        SimPEG.Survey.BaseRx.__init__(self, locs, rxType, **kwargs)

    @property
    def geometric_factor(self):
        return self._geometric_factor

    @property
    def dc_voltage(self):
        return self._dc_voltage

    @property
    def projField(self):
        """Field Type projection (e.g. e b ...)"""
        return self.knownRxTypes[self.rxType][0]

    def projGLoc(self, f):
        """Grid Location projection (e.g. Ex Fy ...)"""
        field = self.knownRxTypes[self.rxType][0]
        comp = self.knownRxTypes[self.rxType][1]
        if comp is not None:
            return f._GLoc(field) + comp
        return f._GLoc(field)

    def eval(self, src, mesh, f):
        P = self.getP(mesh, self.projGLoc(f))
        return P*f[src, self.projField]

    def evalDeriv(self, src, mesh, f, v, adjoint=False):
        P = self.getP(mesh, self.projGLoc(f))
        if not adjoint:
            return P*v
        elif adjoint:
            return P.T*v


# DC.Rx.Dipole(locs)
class Dipole(BaseRx):
    """
    Dipole receiver
    """

    # Threshold to be assumed as a pole receiver
    threshold = 1e-5

    def __init__(self, locsM, locsN, rxType='phi', **kwargs):
        assert locsM.shape == locsN.shape, ('locsM and locsN need to be the '
                                            'same size')
        locs = [np.atleast_2d(locsM), np.atleast_2d(locsN)]
        # We may not need this ...
        BaseRx.__init__(self, locs, rxType)

    @property
    def nD(self):
        """Number of data in the receiver."""
        return self.locs[0].shape[0]

    def getP(self, mesh, Gloc):
        if mesh in self._Ps:
            return self._Ps[mesh]

        # Find indices for pole receivers
        inds_dipole = (
            np.linalg.norm(self.locs[0]-self.locs[1], axis=1) > self.threshold
        )

        P0 = mesh.getInterpolationMat(self.locs[0][inds_dipole], Gloc)
        P1 = mesh.getInterpolationMat(self.locs[1][inds_dipole], Gloc)
        P = P0 - P1

        # Generate interpolation matrix for pole receivers
        if ~np.alltrue(inds_dipole):
            P0_pole = mesh.getInterpolationMat(
                self.locs[0][~inds_dipole], Gloc
            )
            P = sp.vstack((P, P0_pole))

        if self.data_type == 'apparent_resistivity':
            P = sdiag(1./self.geometric_factor) * P
        elif self.data_type == 'apparent_chargeability':
            P = sdiag(1./self.dc_voltage) * P

        if self.storeProjections:
            self._Ps[mesh] = P

        return P


class Dipole_ky(BaseRx):
    """
    Dipole receiver for 2.5D simulations
    """

    def __init__(self, locsM, locsN, rxType='phi', **kwargs):
        assert locsM.shape == locsN.shape, ('locsM and locsN need to be the '
                                            'same size')
        locs = [np.atleast_2d(locsM), np.atleast_2d(locsN)]
        # We may not need this ...
        BaseRx.__init__(self, locs, rxType)

    @property
    def nD(self):
        """Number of data in the receiver."""
        return self.locs[0].shape[0]

    def getP(self, mesh, Gloc):
        if mesh in self._Ps:
            return self._Ps[mesh]

        P0 = mesh.getInterpolationMat(self.locs[0], Gloc)
        P1 = mesh.getInterpolationMat(self.locs[1], Gloc)
        P = P0 - P1

        if self.data_type == 'apparent_resistivity':
            P = sdiag(1./self.geometric_factor) * P
        elif self.data_type == 'apparent_chargeability':
            P = sdiag(1./self.dc_voltage) * P
        if self.storeProjections:
            self._Ps[mesh] = P
        return P

    def eval(self, kys, src, mesh, f):
        P = self.getP(mesh, self.projGLoc(f))
        Pf = P*f[src, self.projField, :]
        return IntTrapezoidal(kys, Pf, y=0.)

    def evalDeriv(self, ky, src, mesh, f, v, adjoint=False):
        P = self.getP(mesh, self.projGLoc(f))
        if not adjoint:
            return P*v
        elif adjoint:
            return P.T*v


class Pole(BaseRx):
    """
    Pole receiver
    """

    def __init__(self, locsM, rxType='phi', **kwargs):

        locs = np.atleast_2d(locsM)
        # We may not need this ...
        BaseRx.__init__(self, locs, rxType)

    @property
    def nD(self):
        """Number of data in the receiver."""
        return self.locs.shape[0]

    def getP(self, mesh, Gloc):
        if mesh in self._Ps:
            return self._Ps[mesh]

        P = mesh.getInterpolationMat(self.locs, Gloc)

        if self.data_type == 'apparent_resistivity':
            P = sdiag(1./self.geometric_factor) * P
        elif self.data_type == 'apparent_chargeability':
            P = sdiag(1./self.dc_voltage) * P
        if self.storeProjections:
            self._Ps[mesh] = P

        return P


class Pole_ky(BaseRx):
    """
    Pole receiver for 2.5D simulations
    """

    def __init__(self, locsM, rxType='phi', **kwargs):

        locs = np.atleast_2d(locsM)
        # We may not need this ...
        BaseRx.__init__(self, locs, rxType)

    @property
    def nD(self):
        """Number of data in the receiver."""
        return self.locs.shape[0]

    def getP(self, mesh, Gloc):
        if mesh in self._Ps:
            return self._Ps[mesh]

        P = mesh.getInterpolationMat(self.locs, Gloc)

        if self.data_type == 'apparent_resistivity':
            P = sdiag(1./self.geometric_factor) * P
        elif self.data_type == 'apparent_chargeability':
            P = sdiag(1./self.dc_voltage) * P
        if self.storeProjections:
            self._Ps[mesh] = P

        return P

    def eval(self, kys, src, mesh, f):
        P = self.getP(mesh, self.projGLoc(f))
        Pf = P*f[src, self.projField, :]
        return IntTrapezoidal(kys, Pf, y=0.)

    def evalDeriv(self, ky, src, mesh, f, v, adjoint=False):
        P = self.getP(mesh, self.projGLoc(f))
        if not adjoint:
            return P*v
        elif adjoint:
            return P.T*v



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import SimPEG
from SimPEG.Utils import Identity, Zero
import numpy as np


class Fields_ky(SimPEG.Problem.TimeFields):

    """

    Fancy Field Storage for a 2.5D code.

    u[:,'phi', kyInd] = phi
    print(u[src0,'phi'])

    Only one field type is stored for
    each problem, the rest are computed. The fields obejct acts like an array
    and is indexed by

    .. code-block:: python

        f = problem.fields(m)
        e = f[srcList,'e']
        j = f[srcList,'j']

    If accessing all sources for a given field, use the

    .. code-block:: python

        f = problem.fields(m)
        phi = f[:,'phi']
        e = f[:,'e']
        b = f[:,'b']

    The array returned will be size (nE or nF, nSrcs :math:`\\times`
    nFrequencies)

    """

    knownFields = {}
    dtype = float

    def _phiDeriv(self, kyInd, src, du_dm_v, v, adjoint=False):
        if (
            getattr(self, '_phiDeriv_u', None) is None or
            getattr(self, '_phiDeriv_m', None) is None
        ):
            raise NotImplementedError(
                'Getting phiDerivs from {0!s} is not '
                'implemented'.format(self.knownFields.keys()[0])
            )

        if adjoint:
            return (self._phiDeriv_u(kyInd, src, v, adjoint=adjoint),
                    self._phiDeriv_m(kyInd, src, v, adjoint=adjoint))

        return (np.array(self._phiDeriv_u(kyInd, src, du_dm_v, adjoint) +
                         self._phiDeriv_m(kyInd, src, v, adjoint),
                         dtype=float))

    def _eDeriv(self, kyInd, src, du_dm_v, v, adjoint=False):
        if (
            getattr(self, '_eDeriv_u', None) is None or
            getattr(self, '_eDeriv_m', None) is None
        ):
            raise NotImplementedError(
                'Getting eDerivs from {0!s} is not '
                'implemented'.format(self.knownFields.keys()[0])
            )

        if adjoint:
            return (self._eDeriv_u(kyInd, src, v, adjoint),
                    self._eDeriv_m(kyInd, src, v, adjoint))
        return (np.array(self._eDeriv_u(kyInd, src, du_dm_v, adjoint) +
                         self._eDeriv_m(kyInd, src, v, adjoint), dtype=float))

    def _jDeriv(self, kyInd, src, du_dm_v, v, adjoint=False):
        if (
            getattr(self, '_jDeriv_u', None) is None or
            getattr(self, '_jDeriv_m', None) is None
        ):
            raise NotImplementedError(
                'Getting jDerivs from {0!s} is not '
                'implemented'.format(self.knownFields.keys()[0])
            )

        if adjoint:
            return (self._jDeriv_u(kyInd, src, v, adjoint),
                    self._jDeriv_m(kyInd, src, v, adjoint))
        return (np.array(self._jDeriv_u(kyInd, src, du_dm_v, adjoint) +
                         self._jDeriv_m(kyInd, src, v, adjoint), dtype=float))

    # def _eDeriv(self, tInd, src, dun_dm_v, v, adjoint=False):
    #     if adjoint is True:
    #         return self._eDeriv_u(tInd, src, v, adjoint), self._eDeriv_m(tInd, src, v, adjoint)
    #     return self._eDeriv_u(tInd, src, dun_dm_v) + self._eDeriv_m(tInd, src, v)

    # def _bDeriv(self, tInd, src, dun_dm_v, v, adjoint=False):
    #     if adjoint is True:
    #         return self._bDeriv_u(tInd, src, v, adjoint), self._bDeriv_m(tInd, src, v, adjoint)
    #     return self._bDeriv_u(tInd, src, dun_dm_v) + self._bDeriv_m(tInd, src, v)


class Fields_ky_CC(Fields_ky):
    """
    Fancy Field Storage for a 2.5D cell centered code.
    """

    knownFields = {'phiSolution': 'CC'}
    aliasFields = {
        'phi': ['phiSolution', 'CC', '_phi'],
        'j': ['phiSolution', 'F', '_j'],
        'e': ['phiSolution', 'F', '_e'],
    }
    # primary - secondary
    # CC variables

    def __init__(self, mesh, survey, **kwargs):
        Fields_ky.__init__(self, mesh, survey, **kwargs)

    def startup(self):
        self.prob = self.survey.prob

    def _GLoc(self, fieldType):
        if fieldType == 'phi':
            return 'CC'
        elif fieldType == 'e' or fieldType == 'j':
            return 'F'
        else:
            raise Exception('Field type must be phi, e, j')

    def _phi(self, phiSolution, src, kyInd):
        return phiSolution

    def _phiDeriv_u(self, kyInd, src, v, adjoint=False):
        return Identity()*v

    def _phiDeriv_m(self, kyInd, src, v, adjoint=False):
        return Zero()

    def _j(self, phiSolution, srcList):
        raise NotImplementedError

    def _e(self, phiSolution, srcList):
        raise NotImplementedError


class Fields_ky_N(Fields_ky):
    """
    Fancy Field Storage for a 2.5D nodal code.
    """
    knownFields = {'phiSolution': 'N'}
    aliasFields = {
        'phi': ['phiSolution', 'N', '_phi'],
        'j': ['phiSolution', 'E', '_j'],
        'e': ['phiSolution', 'E', '_e'],
    }
    # primary - secondary
    # CC variables

    def __init__(self, mesh, survey, **kwargs):
        Fields_ky.__init__(self, mesh, survey, **kwargs)

    def startup(self):
        self.prob = self.survey.prob

    def _GLoc(self, fieldType):
        if fieldType == 'phi':
            return 'N'
        elif fieldType == 'e' or fieldType == 'j':
            return 'E'
        else:
            raise Exception('Field type must be phi, e, j')

    def _phi(self, phiSolution, src, kyInd):
        return phiSolution

    def _phiDeriv_u(self, kyInd, src, v, adjoint=False):
        return Identity()*v

    def _phiDeriv_m(self, kyInd, src, v, adjoint=False):
        return Zero()

    def _j(self, phiSolution, srcList):
        raise NotImplementedError

    def _e(self, phiSolution, srcList):
        raise NotImplementedError



from __future__ import print_function

from discretize.utils.interputils import interpmat

from .matutils import (
    mkvc, sdiag, sdInv, speye, kron3, spzeros, ddx, av,
    av_extrap, ndgrid, ind2sub, sub2ind, getSubArray,
    inv3X3BlockDiagonal, inv2X2BlockDiagonal, TensorType,
    makePropertyTensor, invPropertyTensor, diagEst, Zero,
    Identity, uniqueRows
)
from .codeutils import (
    memProfileWrapper, hook, setKwargs,
    printTitles, printLine, checkStoppers, printStoppers,
    callHooks, dependentProperty,
    asArray_N_x_Dim, requires
)
from .meshutils import (
    exampleLrmGrid, meshTensor, closestPoints, ExtractCoreMesh
)
from .curvutils import volTetra, faceInfo, indexCube
from .CounterUtils import Counter, count, timeIt
from . import ModelBuilder
from . import SolverUtils
from .coordutils import rotatePointsFromNormals, rotationMatrixFromNormals
from .modelutils import surface2ind_topo
from .PlotUtils import plot2Ddata, plotLayer
from .io_utils import download

from .printinfo import Versions



from __future__ import print_function
# Analytic solution of EM fields due to a plane wave

import numpy as np, SimPEG as simpeg
from scipy.constants import mu_0, epsilon_0 as eps_0

def getEHfields(m1d,sigma,freq,zd,scaleUD=True,scaleValue=1):
    '''Analytic solution for MT 1D layered earth. Returns E and H fields.

    :param SimPEG.mesh, object m1d: Mesh object with the 1D spatial information.
    :param numpy.ndarray, vector sigma: Physical property of conductivity corresponding with the mesh.
    :param float, freq: Frequency to calculate data at.
    :param numpy array, vector zd: location to calculate EH fields at
    :param boolean, scaleUD: scales the output to be scaleValue at the top, increases numerical stability.

    Assumes a halfspace with the same conductive as the deepest cell.

    '''
    # Note add an error check for the mesh and sigma are the same size.

    # Constants: Assume constant
    mu = mu_0*np.ones((m1d.nC+1))
    eps = eps_0*np.ones((m1d.nC+1))
    # Angular freq
    w = 2*np.pi*freq
    # Add the halfspace value to the property
    sig = np.concatenate((np.array([sigma[0]]),sigma))
    # Calculate the wave number
    k = np.sqrt(eps*mu*w**2-1j*mu*sig*w)

    # Initiate the propagation matrix, in the order down up.
    UDp = np.zeros((2,m1d.nC+1),dtype=complex)
    UDp[1,0] = scaleValue # Set the wave amplitude as 1 into the half-space at the bottom of the mesh
    # Loop over all the layers, starting at the bottom layer
    for lnr, h in enumerate(m1d.hx): # lnr-number of layer, h-thickness of the layer
        # Calculate
        yp1 = k[lnr]/(w*mu[lnr]) # Admittance of the layer below the current layer
        zp = (w*mu[lnr+1])/k[lnr+1] # Impedance in the current layer
        # Build the propagation matrix

        # Convert fields to down/up going components in layer below current layer
        Pj1 = np.array([[1,1],[yp1,-yp1]],dtype=complex)
        # Convert fields to down/up going components in current layer
        Pjinv = 1./2*np.array([[1,zp],[1,-zp]],dtype=complex)
        # Propagate down and up components through the current layer
        elamh = np.array([[np.exp(-1j*k[lnr+1]*h),0],[0,np.exp(1j*k[lnr+1]*h)]])

        # The down and up component in current layer.
        UDp[:,lnr+1] = elamh.dot(Pjinv.dot(Pj1)).dot(UDp[:,lnr])

        if scaleUD:
            # Scale the values such that 1 at the top
            scaleVal = UDp[:,lnr+1::-1]/UDp[1,lnr+1]
            if np.any(~np.isfinite(scaleVal)):
                # If there is a nan (thickness very great), rebuild the move up cell
                scaleVal = np.zeros_like(UDp[:,lnr+1::-1],dtype=complex)
                scaleVal[1,0] = scaleValue

            UDp[:,lnr+1::-1] = scaleVal

    # Calculate the fields
    Ed = np.empty((zd.size,),dtype=complex)
    Eu = np.empty((zd.size,),dtype=complex)
    Hd = np.empty((zd.size,),dtype=complex)
    Hu = np.empty((zd.size,),dtype=complex)

    # Loop over the layers and calculate the fields
    # In the halfspace below the mesh
    dup = m1d.vectorNx[0]
    dind = dup >= zd
    Ed[dind] = UDp[1,0]*np.exp(-1j*k[0]*(dup-zd[dind]))
    Eu[dind] = UDp[0,0]*np.exp(1j*k[0]*(dup-zd[dind]))
    Hd[dind] = (k[0]/(w*mu[0]))*UDp[1,0]*np.exp(-1j*k[0]*(dup-zd[dind]))
    Hu[dind] = -(k[0]/(w*mu[0]))*UDp[0,0]*np.exp(1j*k[0]*(dup-zd[dind]))
    for ki,mui,epsi,dlow,dup,Up,Dp in zip(k[1::],mu[1::],eps[1::],m1d.vectorNx[:-1],m1d.vectorNx[1::],UDp[0,1::],UDp[1,1::]):
        dind = np.logical_and(dup >= zd, zd > dlow)
        Ed[dind] = Dp*np.exp(-1j*ki*(dup-zd[dind]))
        Eu[dind] = Up*np.exp(1j*ki*(dup-zd[dind]))
        Hd[dind] = (ki/(w*mui))*Dp*np.exp(-1j*ki*(dup-zd[dind]))
        Hu[dind] = -(ki/(w*mui))*Up*np.exp(1j*ki*(dup-zd[dind]))

    # Return return the fields
    return Ed, Eu, Hd, Hu

def getImpedance(m1d,sigma,freq):
    """Analytic solution for MT 1D layered earth. Returns the impedance at the surface.

    :param SimPEG.mesh, object m1d: Mesh object with the 1D spatial information.
    :param numpy.ndarray, vector sigma: Physical property corresponding with the mesh.
    :param numpy.ndarray, vector freq: Frequencies to calculate data at.


    """

    # Initiate the impedances
    Z1d = np.empty(len(freq) , dtype='complex')
    h = m1d.hx   #vectorNx[:-1]
    # Start the process
    for nrFr, fr in enumerate(freq):
        om = 2*np.pi*fr
        Zall = np.empty(len(h)+1,dtype='complex')
        # Calculate the impedance for the bottom layer
        Zall[0] = (mu_0*om)/np.sqrt(mu_0*eps_0*(om)**2 - 1j*mu_0*sigma[0]*om)

        for nr,hi in enumerate(h):
            # Calculate the wave number
            # print(nr,sigma[nr])
            k = np.sqrt(mu_0*eps_0*om**2 - 1j*mu_0*sigma[nr]*om)
            Z = (mu_0*om)/k

            Zall[nr+1] = Z *((Zall[nr] + Z*np.tanh(1j*k*hi))/(Z + Zall[nr]*np.tanh(1j*k*hi)))

        #pdb.set_trace()
        Z1d[nrFr] = Zall[-1]

    return Z1d



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import SimPEG
from SimPEG.Utils import Zero, closestPoints, mkvc
import numpy as np


class BaseSrc(SimPEG.Survey.BaseSrc):
    """
    Base DC source
    """

    current = 1.0
    loc = None
    _q = None

    def __init__(self, rxList, **kwargs):
        SimPEG.Survey.BaseSrc.__init__(self, rxList, **kwargs)

    def eval(self, prob):
        raise NotImplementedError

    def evalDeriv(self, prob):
        return Zero()


class Dipole(BaseSrc):
    """
    Dipole source
    """

    def __init__(self, rxList, locA, locB, **kwargs):
        assert locA.shape == locB.shape, ('Shape of locA and locB should be '
                                          'the same')
        self.loc = [locA, locB]
        BaseSrc.__init__(self, rxList, **kwargs)

    def eval(self, prob):
        if self._q is not None:
            return self._q
        else:
            if prob._formulation == 'HJ':
                inds = closestPoints(prob.mesh, self.loc, gridLoc='CC')
                self._q = np.zeros(prob.mesh.nC)
                self._q[inds] = self.current * np.r_[1., -1.]
            elif prob._formulation == 'EB':
                qa = prob.mesh.getInterpolationMat(self.loc[0],
                                                   locType='N').toarray()
                qb = -prob.mesh.getInterpolationMat(self.loc[1],
                                                    locType='N').toarray()
                self._q = self.current * (qa+qb)
            return self._q


class Pole(BaseSrc):

    def __init__(self, rxList, loc, **kwargs):
        BaseSrc.__init__(self, rxList, loc=loc, **kwargs)

    def eval(self, prob):
        if self._q is not None:
            return self._q
        else:
            if prob._formulation == 'HJ':
                inds = closestPoints(prob.mesh, self.loc)
                self._q = np.zeros(prob.mesh.nC)
                self._q[inds] = self.current * np.r_[1.]
            elif prob._formulation == 'EB':
                q = prob.mesh.getInterpolationMat(
                    self.loc, locType='N'
                )
                self._q = self.current * q.toarray()
            return self._q
